# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

from burnman.eos.birch_murnaghan import BirchMurnaghanBase as BM3
from burnman.eos.mie_grueneisen_debye import MGDBase
from burnman.classes.calibrant import Calibrant
import numpy as np


"""
Zha_2008
^^^^^^^^
"""


class Pt(Calibrant):
    """
    The Pt pressure standard reported by
    Zha (2008; https://doi.org/10.1063/1.2844358).
    """

    def __init__(self):
        def _pressure_Zha_Pt(volume, temperature, params):

            # Isothermal pressure (GPa)
            pressure_model = BM3()
            P0 = pressure_model.pressure(params["T_0"], volume, params)

            # Thermal pressure
            thermal_model = MGDBase()
            Pth0 = thermal_model._thermal_pressure(params["T_0"], volume, params)
            Pth = thermal_model._thermal_pressure(temperature, volume, params)

            # Electronic pressure
            Pel = (
                1.1916e-15 * temperature**4.0
                - 1.4551e-11 * temperature**3.0
                + 1.6209e-07 * temperature**2.0
                + 1.8269e-4 * temperature
                - 0.069
            ) * 1.0e09

            # Total pressure
            P = P0 + Pth - Pth0 + Pel

            return P

        _params_Zha_Pt = {
            "V_0": 9.0904e-06,
            "K_0": 273.5e9,
            "Kprime_0": 4.7,
            "Debye_0": 230.0,  # 370-405
            "grueneisen_0": 2.75,
            "q_0": 0.25,
            "n": 1.0,
            "T_0": 300.0,
            "P_0": 0.0,
            "Z": 4.0,
        }

        Calibrant.__init__(self, _pressure_Zha_Pt, "pressure", _params_Zha_Pt)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.


# This is a script that converts a version of the
# HeFESTo data format into the standard burnman
# format (printed to stdout)
# This script is used to generate the 2024 dataset,
# taking parameters from https://github.com/stixrude/HeFESTo_parameters_010123
# via git clone https://github.com/stixrude/HeFESTo_parameters_010123.git
# The SLB_2024 BurnMan file is then created via the commands
# python HeFESTo_2024_to_burnman.py > ../../minerals/SLB_2024.py; black ../../minerals/SLB_2024.py
# and from . import SLB_2024 added to the minerals __init__.py


import os
import pathlib
import pandas as pd
import numpy as np
from burnman.utils.chemistry import dictionarize_formula
from burnman.constants import gas_constant
from itertools import groupby
import re


def all_equal(iterable):
    g = groupby(iterable)
    return next(g, True) and not next(g, False)


def rfloat(x, m=1.0):
    return round(float(x) * m, ndigits=10)


# Hard-coded into HeFESTo
"""
Ferropericlase:
Crystallography of wustites shows that the locations of the tetrahedral site
and the octahedral vacancies are highly correlated, with the tetrahedral
site being surrounded by a tetrahedral array of octahedral vacancies
(Gavarri et al. 1979 ).Because of this highly structured arrangement,
we ignore the contribution to the ideal entropy of mixing from the
tetrahedral site and the vacant octahedral sites.

...the formula of mag : [Fe]Fe2O4, with the square brackets indicating
that the first site does not participate in mixing. Thus, the ideal entropy
of mixing of an equimolar solution of wu and mag is 2Rln2 per 4 oxygens,
where R is the gas constant.

Spinel:
Pure magnetite is known to have an inverse arrangement of cations at
ambient conditions, rather than normal, as we have assumed.

Bridgmanite:
hepv and hlpv where only the cation on the
smaller B site undergoes the high spin to low spin transition...
"""

hardcoded_site_occupancies = {
    "wu": "[Fe]2[Fe]2",
    "wuls": "[Fels]2[Fels]2",
    "mag": "[Fef]0[Fef]2",  # first site not counted
    "smag": "[Fe][Fef][Fef]",
    "hepv": "[Fef][Fef]",
    "hlpv": "[Fef][Fefls]",
    "fapv": "[Fef][Al]",
    "hppv": "[Fef][Fef]",
    "hmag": "[Fe][Fef][Fef]",
    "acm": "[Na][Fef][Si]2",
    "andr": "[Ca]3[Fef][Fef]",
    "hem": "[Fef][Fef]",
    "lppv": "[Fefls][Fefls]",
}

magnetic_structural_parameters = {
    "fea": {"p": 0.4},
    "mag": {"p": 0.4},
    "smag": {"p": 0.4},
    "hmag": {"p": 0.4},
}

hefesto_path = "HeFESTo_parameters_010123"
mbrdir = pathlib.Path(hefesto_path)
phasedir = pathlib.Path(f"{hefesto_path}/phase")

param_dict = {}
with open("hefesto_parameter_names.txt") as file:
    lines = [line.rstrip() for line in file]
    for line in lines:
        param_dict[" ".join(line.split()[1:])] = line.split()[0]

ignored_mbrs = []

solution_aliases = {
    "c2c": "c2c_pyroxene",
    "cf": "calcium_ferrite_structured_phase",
    "cpx": "clinopyroxene",
    "gt": "garnet",
    "il": "ilmenite",
    "mw": "ferropericlase",
    "nal": "new_aluminous_phase",
    "ol": "olivine",
    "opx": "orthopyroxene",
    "plg": "plagioclase",
    "ppv": "post_perovskite",
    "pv": "bridgmanite",
    "ri": "ringwoodite",
    "sp": "mg_fe_aluminous_spinel",
    "wa": "wadsleyite",
}

# solutions
sol_params = {}
for f in phasedir.iterdir():
    # Do not process liquid!!
    if os.stat(f).st_size > 0 and str(f).split("/")[-1] != "liq":
        name = str(f).split("/")[-1]
        d = pd.read_csv(f, sep="\\s+")

        endmembers = list(d.columns)

        n_mbrs = len(endmembers)

        data = d.values.tolist()

        n_data = len(data)
        v0 = n_data - n_mbrs

        # Preprocess interaction matrices
        # to get rid of ignored endmembers
        use_indices = np.array(
            [i for i, mbr in enumerate(endmembers) if mbr not in ignored_mbrs]
        )
        use_endmembers = np.array(
            [mbr for mbr in endmembers if mbr not in ignored_mbrs]
        )
        n_use_mbrs = len(use_indices)

        energy_matrix = np.array(
            np.array(data)[np.ix_(use_indices, use_indices)], dtype=float
        )
        volume_matrix = np.array(
            np.array(data)[np.ix_(v0 + use_indices, use_indices)], dtype=float
        )

        if n_data == 2 * n_mbrs + 1:
            energy_interactions = []
            energy_interactions_2 = []
            volume_interactions = []
            for i in range(n_use_mbrs - 1):
                j = i + 1
                energy_interactions.append(
                    list(np.around(energy_matrix[i][j:] * 1000.0, decimals=6))
                )
                energy_interactions_2.append(
                    list(np.around(energy_matrix[i][:j] * 1000.0, decimals=6))
                )
                volume_interactions.append(
                    list(np.around(volume_matrix[i][j:] * 1.0e-6, decimals=20))
                )

            max_abs_e = np.max(
                np.abs([item for sublist in energy_interactions for item in sublist])
            )
            max_abs_v = np.max(
                np.abs([item for sublist in volume_interactions for item in sublist])
            )
            max_abs_lower_triangular_e2 = np.max(
                np.abs([item for sublist in energy_interactions_2 for item in sublist])
            )
            assert max_abs_lower_triangular_e2 < 1.0e-3

            # print(endmembers, n_mbrs)
            # print(energy_interactions)
            if max_abs_v > 1.0e-20:
                # print(volume_interactions)
                pass

            alphas = []
            mbr_sites = []
            for mbr in use_endmembers:
                with open(f"{hefesto_path}/{mbr}") as file:
                    lines = [line.rstrip() for line in file]
                    i_laar = [i for i, line in enumerate(lines) if "Laar" in line][0]
                    alphas.append(float(lines[i_laar].split()[0]))

                    formula_raw = lines[0].split()[0]

                    # Catch for MgTs mixing
                    if mbr == "mgts":
                        formula_raw = "Mg_1Al_1Si_2O_6"

                    # Catch for plag mixing
                    if mbr == "ab":
                        formula_raw = "Na_1Al_2Si_2O_8"

                    # Catch for mt mixing
                    if mbr == "mag":
                        formula_raw = "(Vac_1Fe_1)Fe_2O_4"

                    starts = [i for i, char in enumerate(formula_raw) if char.isupper()]
                    mixstarts = [i for i, char in enumerate(formula_raw) if char == "("]
                    mixends = [i for i, char in enumerate(formula_raw) if char == ")"]

                    for i in range(len(mixstarts)):
                        ms = mixstarts[i]
                        me = mixends[i]
                        starts = [i for i in starts if i < ms or i > me]

                    starts.extend(mixstarts)
                    starts.sort()
                    sites = [
                        formula_raw[i:j] for i, j in zip(starts, starts[1:] + [None])
                    ]
                    mbr_sites.append(sites)

            n_sites = [len(sites) for sites in mbr_sites]

            assert all_equal(n_sites)

            active_sites = []
            for i in range(n_sites[0]):
                occupancies = [sites[i] for sites in mbr_sites]
                if not all_equal(occupancies):
                    active_sites.append(i)

            site_formula_strings = []
            totals = []
            for mbr_site in mbr_sites:
                totals.append([])
                site_formula_strings.append("")
                for i in active_sites:
                    # Get number of atoms on sites
                    n_atoms = [int(s) for s in re.findall(r"\d+", mbr_site[i])]

                    species_starts = [
                        i for i, char in enumerate(mbr_site[i]) if char.isupper()
                    ]
                    species_ends = [
                        i for i, char in enumerate(mbr_site[i]) if char == "_"
                    ]
                    total_atoms = np.sum(n_atoms)
                    totals[-1].append(total_atoms)

                    site_string = "["
                    if len(n_atoms) > 1:
                        species = ""
                        for j in range(len(n_atoms)):
                            species += mbr_site[i][species_starts[j] : species_ends[j]]
                            species += f"{n_atoms[j]}/{total_atoms}"
                        site_string += species
                    else:
                        site_string += mbr_site[i][species_starts[0] : species_ends[0]]

                    site_string += "]"
                    if total_atoms != 1:
                        site_string += f"{total_atoms}"

                    site_formula_strings[-1] += site_string

            for i in range(len(totals[0])):
                t = [total[i] for total in totals]
                assert all_equal(t)

            solution_is_symmetric = np.max(np.abs(np.array(alphas) - 1.0)) < 1.0e-5

            sol_params[name] = {
                "name": name,
                "n_mbrs": n_use_mbrs,
                "mbr_names": use_endmembers,
                "mbr_site_formulae": site_formula_strings,
                "alphas": alphas,
            }

            if max_abs_e < 1.0:
                sol_params[name]["solution_type"] = "IdealSolution"
            elif solution_is_symmetric:
                sol_params[name]["solution_type"] = "SymmetricRegularSolution"
            else:
                sol_params[name]["solution_type"] = "AsymmetricRegularSolution"

            if max_abs_e > 1.0:
                sol_params[name]["energy_interaction"] = energy_interactions
            if max_abs_v > 1.0e-20:
                sol_params[name]["volume_interaction"] = volume_interactions

        else:
            pass

# endmembers
mbr_params = {}
mbr_mods = {}
mbr_names = {}
for f in mbrdir.iterdir():
    if os.stat(f).st_size > 0:
        name = str(f).split("/")[-1]

        process = False
        try:
            with open(f) as file:
                lines = [line.rstrip().split() for line in file]
                if lines[-1][-1] == "prime":
                    process = True
                if lines[0][-1] == "Liquid":
                    process = False
        except IsADirectoryError:
            pass

    if process:
        form = lines[0][0].replace("_", "").replace("(", "").replace(")", "")
        formula = dictionarize_formula(form)
        full_name = "_".join(lines[0][1:])
        # formula = formula_to_string(formula)
        # print(formula)
        # print(name)
        idict = {}
        for line in lines[1:]:
            key = param_dict[" ".join(line[1:])]
            idict[key] = line[0]

        if "T_crit" not in idict:
            idict["T_crit"] = "0.00000"

        assert idict["debye"] == "1.00000"

    if idict["zpp"] != "0.00000":
        process = False

    if process:
        mbr_params[name] = {
            "name": full_name,
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": rfloat(idict["F_0"], 1.0e3),
            "V_0": rfloat(idict["V_0"], 1.0e-6),
            "K_0": rfloat(idict["K_0"], 1.0e9),
            "Kprime_0": rfloat(idict["Kprime_0"]),
            "Debye_0": rfloat(idict["Theta_0"]),
            "grueneisen_0": rfloat(idict["gamma_0"]),
            "q_0": rfloat(idict["q_0"]),
            "G_0": rfloat(idict["G_0"], 1.0e9),
            "Gprime_0": rfloat(idict["Gprime_0"]),
            "eta_s_0": rfloat(idict["eta_s_0"]),
            "n": rfloat(idict["n"]),
            "Z": rfloat(idict["Z"]),
            "molar_mass": rfloat(idict["molar_mass"], 1.0e-3),
        }

        # electronic terms
        if np.abs(rfloat(idict["bel_0"])) > 1.0e-10:
            mbr_params[name]["equation_of_state"] = "slb3-conductive"
            mbr_params[name]["bel_0"] = rfloat(idict["bel_0"])
            mbr_params[name]["gel"] = rfloat(idict["gel"])

        # The positivity constraint for T_crit below
        # is hardcoded into HeFESTo.
        if idict["T_crit"] != "0.00000" and float(idict["T_crit"]) > 0.0:
            if name in magnetic_structural_parameters:
                # The magnetic model in SLB2024 is identical
                # to the chs model already coded in BurnMan
                # but with the zero Gibbs energy and entropy
                # assigned to the fully ordered form,
                # not the disordered form,.
                p = magnetic_structural_parameters[name]["p"]
                Scrit = float(idict["S_crit"])
                magnetic_moment = np.exp(Scrit / gas_constant) - 1.0
                A = (518.0 / 1125.0) + (11692.0 / 15975.0) * ((1.0 / p) - 1.0)
                fT0 = (1.0 / A) * float(idict["T_crit"]) * 79.0 / (140.0 * p)
                E0 = Scrit * fT0

                mbr_mods[name] = [
                    [
                        "magnetic_chs",
                        {
                            "structural_parameter": p,
                            "curie_temperature": [float(idict["T_crit"]), 0.0],
                            "magnetic_moment": [magnetic_moment, 0.0],
                        },
                    ],
                    [
                        "linear",
                        {
                            "delta_E": E0,
                            "delta_S": Scrit,
                            "delta_V": 0.0,
                        },
                    ],
                ]
            else:
                mbr_mods[name] = [
                    [
                        "landau_slb_2022",
                        {
                            "Tc_0": float(idict["T_crit"]),
                            "S_D": float(idict["S_crit"]),
                            "V_D": float(idict["V_crit"]) / 1.0e6,
                        },
                    ]
                ]
        mbr_names[name] = (
            full_name.lower()
            .replace("-", "_")
            .replace(" ", "_")
            .replace("'", "")
            .replace('\\"', "")
        )

# WRITE FILE

print(
    "# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit\n"
    "# for the Earth and Planetary Sciences\n"
    "# Copyright (C) 2012 - 2024 by the BurnMan team,\n"
    "# released under the GNU GPL v2 or later.\n"
    "\n\n"
    '"""\n'
    "SLB_2024\n"
    "Minerals from Stixrude & Lithgow-Bertelloni 2024 and references therein\n"
    f"File autogenerated using HeFESTO_2024_to_burnman.py and {hefesto_path}\n"
    '"""\n\n'
    "from __future__ import absolute_import\n"
    "\n\n"
    "from ..eos.slb import SLB3\n"
    "from ..eos import debye\n"
    "from ..eos import birch_murnaghan as bm\n"
    "from ..classes.mineral import Mineral\n"
    "from ..classes.solution import Solution, RelaxedSolution\n"
    "from ..classes.solutionmodel import (\n"
    "    IdealSolution,\n"
    "    SymmetricRegularSolution,\n"
    "    AsymmetricRegularSolution,\n"
    ")\n\n"
)


st_class = (
    "class SLB3Stishovite(SLB3):\n"
    "    def shear_modulus(self, pressure, temperature, volume, params):\n"
    '        """\n'
    "        Returns the shear modulus. :math:`[Pa]`\n"
    "        SLB2024 first calculate the shear modulus at T_0 using\n"
    "        the third order Birch-Murnaghan equation of state,\n"
    "        then apply the softening relevant to the temperature\n"
    "        and pressure of interest, and then apply the usual thermal term.\n"
    '        """\n'
    "\n"
    "        G_bare = bm.shear_modulus_third_order(volume, params)\n"
    "        Pc = 51.6e9 + 11.1 * (temperature - 300.0) * 1.0e6  # Eq B4\n"
    "        PcQ = (\n"
    "            Pc + 50.7e9\n"
    "        )  # From the HeFESTo code in stishtran.f. PcQ is T-independent in paper.\n"
    "        Cs0 = 128.0e9 * PcQ / Pc\n"
    "        Delta = 0.0  # 100.e9 in the HeFESTo code in stishtran.f, 0. in paper.\n"
    "        PminusPc = pressure - Pc\n"
    "        if PminusPc < 0.0:\n"
    "            Cs = Cs0 * PminusPc / (pressure - PcQ)  # Eq B2\n"
    "        else:\n"
    "            Ast = (Cs0 - Delta) * (PcQ - Pc)\n"
    "            Cs = Cs0 - Ast / (PcQ - pressure + 3.0 * PminusPc)  # Eq B3\n"
    "\n"
    "        # VRH-like average of the bare shear modulus with softened (C11-C12)/2.\n"
    "        G = 0.5 * (13.0 / 15.0 * G_bare + 2.0 / 15.0 * Cs) + 0.5 / (\n"
    "            13.0 / 15.0 / G_bare + 2.0 / 15.0 / Cs\n"
    "        )  # Eq B5\n"
    "\n"
    "        debye_T = self._debye_temperature(params['V_0'] / volume, params)\n"
    "        eta_s = self._isotropic_eta_s(params['V_0'] / volume, params)\n"
    "\n"
    "        E_th = debye.thermal_energy(temperature, debye_T, params['n'])\n"
    "        E_th_ref = debye.thermal_energy(params['T_0'], debye_T, params['n'])\n"
    "\n"
    "        return G - eta_s * (E_th - E_th_ref) / volume\n"
)

fper_relaxed_class = (
    "class ferropericlase_relaxed(RelaxedSolution):\n"
    "    def __init__(self):\n"
    '        """RelaxedSolution model for ferropericlase (mw).\n'
    "        Endmembers (and site species distributions) are given in the order:\n"
    "        - pe ([Mg]2[Mg]2)\n"
    "        - wu and wuls ([Fe,Fels]2[Fe,Fels]2)\n"
    "        - anao ([Na]2[Al]2)\n"
    "        - mag ([Fe]0[Fef]2)\n"
    "        The entropy from the first site in magnetite is not counted.\n"
    "        The equilibrium spin state is calculated automatically.\n"
    '        """\n'
    "        solution = ferropericlase()\n"
    "        vrel = [[0.0, 0.5, -0.5, 0.0, 0.0]]\n"
    "        vunrel = [\n"
    "            [1.0, 0.0, 0.0, 0.0, 0.0],\n"
    "            [0.0, 0.5, 0.5, 0.0, 0.0],\n"
    "            [0.0, 0.0, 0.0, 1.0, 0.0],\n"
    "            [0.0, 0.0, 0.0, 0.0, 1.0],\n"
    "        ]\n"
    "        RelaxedSolution.__init__(self, solution, vrel, vunrel)\n"
)

bdg_relaxed_class = (
    "class bridgmanite_relaxed(RelaxedSolution):\n"
    "    def __init__(self):\n"
    '        """RelaxedSolution model for bridgmanite (pv).\n'
    "        Only the spin transition is relaxed.\n"
    "        Endmembers (and site species distributions) are given in the order:\n"
    "        - mgpv ([Mg][Si])\n"
    "        - fepv ([Fe][Si])\n"
    "        - alpv ([Al][Al])\n"
    "        - hepv + hlpv ([Fef][Fef,Fefls])\n"
    "        - fapv ([Fef][Al])\n"
    "        - crpv ([Cr][Cr])\n"
    "        The equilibrium spin state is calculated automatically.\n"
    '        """\n'
    "        solution = bridgmanite()\n"
    "        vrel = [[0.0, 0., 0., 0.5, -0.5, 0.0, 0.0]]\n"
    "        vunrel = [\n"
    "            [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],\n"
    "            [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],\n"
    "            [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],\n"
    "            [0.0, 0.0, 0.0, 0.5, 0.5, 0.0, 0.0],\n"
    "            [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0],\n"
    "            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],\n"
    "        ]\n"
    "        RelaxedSolution.__init__(self, solution, vrel, vunrel)\n"
)

print('"""\n' "ENDMEMBERS\n" '"""\n')
for key, mbr_prm in sorted(mbr_params.items()):
    if key == "st":
        print(st_class)
        mbr_prm = mbr_prm.__repr__().replace("'slb3'", "SLB3Stishovite()")
        mbr_prm = mbr_prm.replace("slb3", "fault")
    print(
        f"\nclass {key}(Mineral):\n"
        "    def __init__(self):\n"
        f"        self.params = {mbr_prm}\n"
    )
    if key in mbr_mods:
        print(
            f"        self.property_modifiers = {mbr_mods[key]}",
        )
        print("")
    print("        Mineral.__init__(self)")
    print("")

print('"""\n' "SOLUTIONS\n" '"""\n')

for key, prm in sorted(sol_params.items()):
    for i in range(prm["n_mbrs"]):
        if prm["mbr_names"][i] in hardcoded_site_occupancies:
            prm["mbr_site_formulae"][i] = hardcoded_site_occupancies[
                prm["mbr_names"][i]
            ]

    docstring = '"""'
    docstring += f'{prm["solution_type"]} model for {solution_aliases[key]} ({key}).\n'
    docstring += "Endmembers (and site species distributions) are given in the order:\n"
    for i in range(prm["n_mbrs"]):
        docstring += f'- {prm["mbr_names"][i]} ({prm["mbr_site_formulae"][i]})\n'
    if key == "mw":
        docstring += "The entropy from the first site in magnetite is not counted.\n"
    docstring += '"""'

    print(
        f"\nclass {solution_aliases[key]}(Solution):\n"
        "    def __init__(self, molar_fractions=None):\n"
        f"        {docstring}\n"
        f'        self.name = "{solution_aliases[key]}"\n'
        f"        self.solution_model = {prm['solution_type']}(\n"
        "            endmembers=["
    )
    for i in range(prm["n_mbrs"]):
        print(
            f'                [{prm["mbr_names"][i]}(), "{prm["mbr_site_formulae"][i]}"],'
        )
    print("            ],")

    if prm["solution_type"] == "AsymmetricRegularSolution":
        print(f"            alphas={prm['alphas']},")

    if prm["solution_type"] != "IdealSolution":
        print(f'            energy_interaction={prm["energy_interaction"]},')
        if "volume_interaction" in prm:
            print(f'            volume_interaction={prm["volume_interaction"]},')

    print("        )\n")
    print("        Solution.__init__(self, molar_fractions=molar_fractions)\n")

    if key == "mw":
        print(fper_relaxed_class)
    if key == "pv":
        print(bdg_relaxed_class)


print('\n"""\n' "ENDMEMBER ALIASES\n" '"""\n')

for key, value in sorted(mbr_names.items()):
    print(f"{value.replace('(', '').replace(')', '')} = {key}")

print("")


print('\n"""\n' "SOLUTION ALIASES\n" '"""\n')

for key, value in sorted(solution_aliases.items()):
    if key != "sp":
        print(f"{key} = {value}")



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""

HP_2011_fluids
^^^^^^^^^^^^^^

Fluids from Holland and Powell 2011 and references therein.
CORK parameters are taken from various sources.

CHO gases from Holland and Powell, 1991:
  - ["CO2",304.2,0.0738]
  - ["CH4",190.6,0.0460]
  - ["H2",41.2,0.0211]
  - ["CO",132.9,0.0350]

H2O and S2 from Wikipedia, 2012/10/23:
  - ["H2O",647.096,0.22060]
  - ["S2",1314.00,0.21000]

H2S from ancyclopedia.airliquide.com, 2012/10/23:
  - ["H2S",373.15,0.08937]

NB: Units for cork[i] in Holland and Powell datasets are:
  - a = kJ^2/kbar*K^(1/2)/mol^2: multiply by 1e-2
  - b = kJ/kbar/mol: multiply by 1e-5
  - c = kJ/kbar^1.5/mol: multiply by 1e-9
  - d = kJ/kbar^2/mol: multiply by 1e-13

Individual terms are divided through by P, P, P^1.5, P^2, so:
  - [0][j]: multiply by 1e6
  - [1][j]: multiply by 1e3
  - [2][j]: multiply by 1e3
  - [3][j]: multiply by 1e3
  - cork_P is given in kbar: multiply by 1e8
"""

from __future__ import absolute_import

from ..classes.mineral import Mineral
from ..utils.chemistry import dictionarize_formula, formula_mass


class CO2(Mineral):
    def __init__(self):
        formula = "CO2"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "carbon dioxide",
            "formula": formula,
            "equation_of_state": "cork",
            "cork_params": [
                [5.45963e1, -8.63920e0],
                [9.18301e-1],
                [-3.30558e-2, 2.30524e-3],
                [6.93054e-4, -8.38293e-5],
            ],
            "cork_T": 304.2,
            "cork_P": 0.0738e8,
            "H_0": -393.51e3,
            "S_0": 213.7,
            "Cp": [87.8, -2.644e-3, 706.4e3, -998.9],
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }
        Mineral.__init__(self)


class CH4(Mineral):
    def __init__(self):
        formula = "CH4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "methane",
            "formula": formula,
            "equation_of_state": "cork",
            "cork_params": [
                [5.45963e1, -8.63920e0],
                [9.18301e-1],
                [-3.30558e-2, 2.30524e-3],
                [6.93054e-4, -8.38293e-5],
            ],
            "cork_T": 190.6,
            "cork_P": 0.0460e8,
            "H_0": -74.81e3,
            "S_0": 186.26,
            "Cp": [150.1, 0.002063, 3427700.0, -2650.4],
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }
        Mineral.__init__(self)


class O2(Mineral):
    def __init__(self):
        formula = "O2"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "oxygen",
            "formula": formula,
            "equation_of_state": "cork",
            "cork_params": [
                [5.45963e1, -8.63920e0],
                [9.18301e-1],
                [-3.30558e-2, 2.30524e-3],
                [6.93054e-4, -8.38293e-5],
            ],
            "cork_T": 0.0,
            "cork_P": 1.0e5,
            "H_0": 0.0,
            "S_0": 205.2,
            "Cp": [48.3, -0.000691, 499200.0, -420.7],
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }
        Mineral.__init__(self)


class H2(Mineral):
    def __init__(self):
        formula = "H2"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "hydrogen",
            "formula": formula,
            "equation_of_state": "cork",
            "cork_params": [
                [5.45963e1, -8.63920e0],
                [9.18301e-1],
                [-3.30558e-2, 2.30524e-3],
                [6.93054e-4, -8.38293e-5],
            ],
            "cork_T": 41.2,
            "cork_P": 0.0211e8,
            "H_0": 0.0,
            "S_0": 130.7,
            "Cp": [23.3, 0.004627, 0.0, 76.3],
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }
        Mineral.__init__(self)


class S2(Mineral):
    def __init__(self):
        formula = "S2"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "sulfur",
            "formula": formula,
            "equation_of_state": "cork",
            "cork_params": [
                [5.45963e1, -8.63920e0],
                [9.18301e-1],
                [-3.30558e-2, 2.30524e-3],
                [6.93054e-4, -8.38293e-5],
            ],
            "cork_T": 1314.00,
            "cork_P": 0.21000e8,
            "H_0": 128.54e3,
            "S_0": 231.0,
            "Cp": [37.1, 0.002398, -161000.0, -65.0],
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }
        Mineral.__init__(self)


class H2S(Mineral):
    def __init__(self):
        formula = "H2S"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "hydrogen sulfide",
            "formula": formula,
            "equation_of_state": "cork",
            "cork_params": [
                [5.45963e1, -8.63920e0],
                [9.18301e-1],
                [-3.30558e-2, 2.30524e-3],
                [6.93054e-4, -8.38293e-5],
            ],
            "cork_T": 373.15,
            "cork_P": 0.08937e8,
            "H_0": 128.54e3,
            "S_0": 231.0,
            "Cp": [47.4, 0.010240, 615900.0, -397.8],
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }
        Mineral.__init__(self)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2022 by the BurnMan team, released under the GNU
# GPL v2 or later.


from __future__ import absolute_import

import numpy as np
from sympy import Matrix
from ..classes.combinedmineral import CombinedMineral
from ..classes.solution import Solution
from ..classes.solutionmodel import (
    IdealSolution,
    SymmetricRegularSolution,
    AsymmetricRegularSolution,
    SubregularSolution,
)
from ..utils.chemistry import site_occupancies_to_strings
from ..utils.math import complete_basis


def _decompose_3D_matrix(W):
    """
    Decomposes a 3D matrix W_ijk where E = W_ijk p_i p_j p_k
    into a subregular form where
    E = G_i p_i + WB_ij (1 - p_j + p_i) / 2 + WT_ijk p_i p_j p_k,
    and i < j < k.

    :param W: 3D interaction matrix.
    :type W: numpy.array

    :returns: The 1D array G_i, the 2D upper triangular array WB_ij and
        the ternary terms in a list where each item is in the form [i, j, k, WT_ijk]
    :rtype: tuple
    """

    n_mbrs = len(W)
    # New endmember components
    # W_iii needs to be copied, otherwise just a view onto W
    new_endmember_excesses = np.copy(np.einsum("iii->i", W))

    # Removal of endmember components from 3D representation
    W -= (
        np.einsum(
            "i, j, k->ijk", new_endmember_excesses, np.ones(n_mbrs), np.ones(n_mbrs)
        )
        + np.einsum(
            "i, j, k->ijk", np.ones(n_mbrs), new_endmember_excesses, np.ones(n_mbrs)
        )
        + np.einsum(
            "i, j, k->ijk", np.ones(n_mbrs), np.ones(n_mbrs), new_endmember_excesses
        )
    ) / 3.0

    # Transformed 2D components
    # (i=j, i=k, j=k)
    new_binary_matrix = (
        np.einsum("jki, jk -> ij", W, np.identity(n_mbrs))
        + np.einsum("jik, jk -> ij", W, np.identity(n_mbrs))
        + np.einsum("ijk, jk -> ij", W, np.identity(n_mbrs))
    ).round(decimals=12)

    # Wb is the 3D matrix corresponding to the terms in the binary matrix,
    # such that the two following print statements produce the same answer
    # for a given array of endmember proportions
    Wb = (
        np.einsum("ijk, ij->ijk", W, np.identity(n_mbrs))
        + np.einsum("ijk, jk->ijk", W, np.identity(n_mbrs))
        + np.einsum("ijk, ik->ijk", W, np.identity(n_mbrs))
    )

    # Remove binary component from 3D representation
    # The extra terms are needed because the binary term in the formulation
    # of a subregular solution model given by
    # Helffrich and Wood includes ternary components (the sum_k X_k part)..
    W -= (
        Wb
        + (
            np.einsum("ij, k", new_binary_matrix, np.ones(n_mbrs))
            - np.einsum("ij, ik->ijk", new_binary_matrix, np.identity(n_mbrs))
            - np.einsum("ij, jk->ijk", new_binary_matrix, np.identity(n_mbrs))
        )
        / 2.0
    )

    # Find the 3D components Wijk by adding the elements at
    # the six equivalent positions in the matrix
    new_ternary_terms = []
    for i in range(n_mbrs):
        for j in range(i + 1, n_mbrs):
            for k in range(j + 1, n_mbrs):
                val = (
                    W[i, j, k]
                    + W[j, k, i]
                    + W[k, i, j]
                    + W[k, j, i]
                    + W[j, i, k]
                    + W[i, k, j]
                ).round(decimals=12)
                if np.abs(val) > 1.0e-12:
                    new_ternary_terms.append([i, j, k, val])

    return (new_endmember_excesses, new_binary_matrix, new_ternary_terms)


def _subregular_matrix_conversion(
    new_basis, binary_matrix, ternary_terms=None, endmember_excesses=None
):
    """
    Converts the arrays reguired to describe a subregular solution
    from one endmember basis to another.

    The excess nonconfigurational energies of the subregular solution model
    are described as follows:
    E = G_i p_i + WB_ij (1 - p_j + p_i) / 2 + WT_ijk p_i p_j p_k,
    and i < j < k.

    :param new_basis: The new endmember basis, given as amounts of the old endmembers.
    :type new_basis: 2D numpy array

    :param binary_matrix: The upper triangular matrix WB_ij.
    :type binary_matrix: 2D numpy array

    :param ternary_terms: The ternary terms in a list where each
        item is in the form [i, j, k, WT_ijk]
    :type ternary_terms: list of lists of length 4

    :param endmember_excesses: The array G_i
    :type endmember_excesses: 1D numpy array

    :returns: The 1D array G_i, the 2D upper triangular array WB_ij and
        the ternary terms in a list where each item is in the form [i, j, k, WT_ijk]
    :rtype: tuple
    """
    n_mbrs = len(binary_matrix)
    # Compact 3D representation of original interactions
    W = (
        np.einsum("i, jk -> ijk", np.ones(n_mbrs), binary_matrix)
        + np.einsum("ij, jk -> ijk", binary_matrix, np.identity(n_mbrs))
        - np.einsum("ij, ik -> ijk", binary_matrix, np.identity(n_mbrs))
    ) / 2.0

    # Add endmember components to 3D representation
    if endmember_excesses is not None:
        W += (
            np.einsum(
                "i, j, k->ijk", endmember_excesses, np.ones(n_mbrs), np.ones(n_mbrs)
            )
            + np.einsum(
                "j, i, k->ijk", endmember_excesses, np.ones(n_mbrs), np.ones(n_mbrs)
            )
            + np.einsum(
                "k, i, j->ijk", endmember_excesses, np.ones(n_mbrs), np.ones(n_mbrs)
            )
        ) / 3.0

    # Add ternary values to 3D representation
    if ternary_terms is not None:
        for i, j, k, val in ternary_terms:
            W[i, j, k] += val

    # Transformation to new 3D representation
    A = new_basis.T
    Wn = np.einsum("il, jm, kn, ijk -> lmn", A, A, A, W)

    new_endmember_excesses, new_binary_terms, new_ternary_terms = _decompose_3D_matrix(
        Wn
    )

    return (new_endmember_excesses, new_binary_terms, new_ternary_terms)


def transform_solution_to_new_basis(
    solution,
    new_basis,
    n_mbrs=None,
    solution_name=None,
    endmember_names=None,
    molar_fractions=None,
):
    """
    Transforms a solution model from one endmember basis to another.
    Returns a new Solution object.

    :param solution: The original solution object.
    :type solution: :class:`burnman.Solution` object

    :param new_basis: The new endmember basis, given as amounts of the old endmembers.
    :type new_basis: 2D numpy array

    :param n_mbrs: The number of endmembers in the new solution
        (defaults to the length of new_basis).
    :type n_mbrs: float, optional

    :param solution_name: A name corresponding to the new solution.
    :type solution_name: str, optional

    :param endmember_names: A list corresponding to the names of the new endmembers.
    :type endmember_names: list of str, optional

    :param molar_fractions: Fractions of the new endmembers in the new solution.
    :type molar_fractions: numpy.array, optional

    :returns: The transformed solution.
    :rtype: :class:`burnman.Solution` object
    """
    new_basis = np.array(new_basis)
    if n_mbrs is None:
        n_mbrs, n_all_mbrs = new_basis.shape
    else:
        _, n_all_mbrs = new_basis.shape

    if solution_name is None:
        name = "child solution"
    else:
        name = solution_name

    solution_model = solution.solution_model

    # Use type here to avoid inheritance problems
    solution_type = type(solution_model)
    if solution_type == IdealSolution:
        ESV_modifiers = [[0.0, 0.0, 0.0] for v in new_basis]

    if (
        solution_type == AsymmetricRegularSolution
        or solution_type == SymmetricRegularSolution
    ):
        A = complete_basis(new_basis).T

        old_alphas = solution.solution_model.alphas
        alphas = np.einsum("i, ij", solution.solution_model.alphas, A)
        inv_diag_alphas = np.diag(1.0 / alphas)
        B = np.einsum("ij, jk, kl->il", np.diag(old_alphas), A, inv_diag_alphas)
        alphas = list(alphas[0:n_mbrs])
        Qe = np.einsum("ik, ij, kl->jl", solution.solution_model.We, B, B)
        Qs = np.einsum("ik, ij, kl->jl", solution.solution_model.Ws, B, B)
        Qv = np.einsum("ik, ij, kl->jl", solution.solution_model.Wv, B, B)

        def new_interactions(Q, n_mbrs):
            return [
                [
                    float(
                        (Q[i, j] + Q[j, i] - Q[i, i] - Q[j, j])
                        * (alphas[i] + alphas[j])
                        / 2.0
                    )
                    for j in range(i + 1, n_mbrs)
                ]
                for i in range(n_mbrs - 1)
            ]

        energy_interaction = new_interactions(Qe, n_mbrs)
        entropy_interaction = new_interactions(Qs, n_mbrs)
        volume_interaction = new_interactions(Qv, n_mbrs)

        ESV_modifiers = [
            [Qe[i, i] * alphas[i], Qs[i, i] * alphas[i], Qv[i, i] * alphas[i]]
            for i in range(n_mbrs)
        ]

    elif solution_type == SubregularSolution:
        full_basis = complete_basis(new_basis)

        def new_interactions(W, n_mbrs):
            return [
                [[W[i, j], W[j, i]] for j in range(i + 1, n_mbrs)]
                for i in range(n_mbrs - 1)
            ]

        # N.B. initial endmember_excesses are zero
        Emod, We, ternary_e = _subregular_matrix_conversion(
            full_basis,
            solution.solution_model.We,
            solution.solution_model.ternary_terms_e,
        )
        Smod, Ws, ternary_s = _subregular_matrix_conversion(
            full_basis,
            solution.solution_model.Ws,
            solution.solution_model.ternary_terms_s,
        )
        Vmod, Wv, ternary_v = _subregular_matrix_conversion(
            full_basis,
            solution.solution_model.Wv,
            solution.solution_model.ternary_terms_v,
        )

        energy_interaction = new_interactions(We, n_mbrs)
        entropy_interaction = new_interactions(Ws, n_mbrs)
        volume_interaction = new_interactions(Wv, n_mbrs)

        ESV_modifiers = [[Emod[i], Smod[i], Vmod[i]] for i in range(n_mbrs)]

    else:
        raise Exception(
            "The function to change basis for the "
            "{0} solution type has not yet been "
            "implemented.".format(solution_type)
        )

    # Create site formulae
    new_occupancies = np.array(new_basis).dot(
        solution.solution_model.endmember_occupancies
    )
    new_multiplicities = np.array(new_basis).dot(
        solution.solution_model.site_multiplicities
    )
    site_formulae = site_occupancies_to_strings(
        solution.solution_model.sites, new_multiplicities, new_occupancies
    )

    # Create endmembers
    endmembers = []
    for i, vector in enumerate(new_basis):
        nonzero_indices = np.nonzero(vector)[0]
        if len(nonzero_indices) == 1:
            endmembers.append(
                [solution_model.endmembers[nonzero_indices[0]][0], site_formulae[i]]
            )
        else:
            mbr = CombinedMineral(
                [solution_model.endmembers[idx][0] for idx in nonzero_indices],
                [vector[idx] for idx in nonzero_indices],
                ESV_modifiers[i],
            )
            mbr.params["formula"] = {
                key: value
                for (key, value) in mbr.params["formula"].items()
                if value > 1.0e-12
            }
            endmembers.append([mbr, site_formulae[i]])

    if endmember_names is not None:
        for i in range(n_mbrs):
            endmembers[i][0].params["name"] = endmember_names[i]
            endmembers[i][0].name = endmember_names[i]

    if n_mbrs == 1:
        endmembers[0][0].name = name
        endmembers[0][0].parent = solution
        endmembers[0][0].basis = new_basis
        return endmembers[0][0]
    else:
        if solution_type == IdealSolution:
            new_solution_model = IdealSolution(endmembers=endmembers)
        elif (
            solution_type == SymmetricRegularSolution
            or solution_type == SubregularSolution
        ):
            new_solution_model = type(solution_model)(
                endmembers=endmembers,
                energy_interaction=energy_interaction,
                volume_interaction=volume_interaction,
                entropy_interaction=entropy_interaction,
            )
        else:
            new_solution_model = type(solution_model)(
                endmembers=endmembers,
                energy_interaction=energy_interaction,
                volume_interaction=volume_interaction,
                entropy_interaction=entropy_interaction,
                alphas=alphas,
            )
        new_solution = Solution(
            name=name,
            solution_model=new_solution_model,
            molar_fractions=molar_fractions,
        )
        new_solution.parent = solution
        new_solution.basis = new_basis
        return new_solution



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2015 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import numpy as np
import scipy.optimize as opt
import scipy.integrate as integrate
import warnings

# Try to import the jit from numba.  If it is
# not available, just go with the standard
# python interpreter
try:
    from numba import jit
except ImportError:

    def jit(fn):
        return fn


from . import birch_murnaghan as bm
from . import debye
from . import equation_of_state as eos
from ..utils.math import bracket


@jit(nopython=True)
def _grueneisen_parameter_fast(V_0, volume, gruen_0, q_0):
    """global function with plain parameters so jit will work"""
    x = V_0 / volume
    f = 1.0 / 2.0 * (pow(x, 2.0 / 3.0) - 1.0)
    a1_ii = 6.0 * gruen_0  # EQ 47
    a2_iikk = -12.0 * gruen_0 + 36.0 * gruen_0 * gruen_0 - 18.0 * q_0 * gruen_0  # EQ 47
    nu_o_nu0_sq = 1.0 + a1_ii * f + (1.0 / 2.0) * a2_iikk * f * f  # EQ 41
    return 1.0 / 6.0 / nu_o_nu0_sq * (2.0 * f + 1.0) * (a1_ii + a2_iikk * f)


def _intgroverVdV(V_0, volume, gruen_0, q_0):
    return integrate.quad(
        lambda x: _grueneisen_parameter_fast(V_0, x, gruen_0, q_0) / x, V_0, volume
    )[0]


@jit(nopython=True)
def _delta_pressure(
    x, pressure, temperature, V_0, T_0, Cv, a1_ii, a2_iikk, b_iikk, b_iikkmm
):
    f = 0.5 * (pow(V_0 / x, 2.0 / 3.0) - 1.0)
    nu_o_nu0_sq = 1.0 + a1_ii * f + (1.0 / 2.0) * a2_iikk * f * f  # EQ 41
    gr = 1.0 / 6.0 / nu_o_nu0_sq * (2.0 * f + 1.0) * (a1_ii + a2_iikk * f)

    return (
        (1.0 / 3.0)
        * (pow(1.0 + 2.0 * f, 5.0 / 2.0))
        * ((b_iikk * f) + (0.5 * b_iikkmm * f * f))
        + gr * Cv * (temperature - T_0) / x
        - pressure
    )  # EQ 21


class DKS_S(eos.EquationOfState):
    """
    Base class for the finite strain solid equation of state detailed
    in :cite:`deKoker2013` (supplementary materials).
    """

    def volume_dependent_q(self, x, params):
        """
        Finite strain approximation for :math:`q`, the isotropic volume strain
        derivative of the grueneisen parameter.
        """
        f = 1.0 / 2.0 * (pow(x, 2.0 / 3.0) - 1.0)
        a1_ii = 6.0 * params["grueneisen_0"]  # EQ 47
        a2_iikk = (
            -12.0 * params["grueneisen_0"]
            + 36.0 * pow(params["grueneisen_0"], 2.0)
            - 18.0 * params["q_0"] * params["grueneisen_0"]
        )  # EQ 47
        nu_o_nu0_sq = 1.0 + a1_ii * f + (1.0 / 2.0) * a2_iikk * f * f  # EQ 41
        gr = 1.0 / 6.0 / nu_o_nu0_sq * (2.0 * f + 1.0) * (a1_ii + a2_iikk * f)
        if (
            np.abs(params["grueneisen_0"]) < 1.0e-10
        ):  # avoids divide by zero if grueneisen_0 = 0.
            q = 1.0 / 9.0 * (18.0 * gr - 6.0)
        else:
            q = (
                1.0
                / 9.0
                * (
                    18.0 * gr
                    - 6.0
                    - 1.0
                    / 2.0
                    / nu_o_nu0_sq
                    * (2.0 * f + 1.0)
                    * (2.0 * f + 1.0)
                    * a2_iikk
                    / gr
                )
            )
        return q

    def _isotropic_eta_s(self, x, params):
        """
        Finite strain approximation for :math:`eta_{s0}`, the isotropic shear
        strain derivative of the grueneisen parameter.
        """
        f = 1.0 / 2.0 * (pow(x, 2.0 / 3.0) - 1.0)
        a2_s = -2.0 * params["grueneisen_0"] - 2.0 * params["eta_s_0"]  # EQ 47
        a1_ii = 6.0 * params["grueneisen_0"]  # EQ 47
        a2_iikk = (
            -12.0 * params["grueneisen_0"]
            + 36.0 * pow(params["grueneisen_0"], 2.0)
            - 18.0 * params["q_0"] * params["grueneisen_0"]
        )  # EQ 47
        nu_o_nu0_sq = 1.0 + a1_ii * f + (1.0 / 2.0) * a2_iikk * pow(f, 2.0)  # EQ 41
        gr = 1.0 / 6.0 / nu_o_nu0_sq * (2.0 * f + 1.0) * (a1_ii + a2_iikk * f)
        # EQ 46 NOTE the typo from Stixrude 2005:
        eta_s = -gr - (
            1.0 / 2.0 * pow(nu_o_nu0_sq, -1.0) * pow((2.0 * f) + 1.0, 2.0) * a2_s
        )

        return eta_s

    # calculate isotropic thermal pressure, see
    # Matas et. al. (2007) eq B4
    def _thermal_pressure(self, T, V, params):
        gr = self.grueneisen_parameter(0.0, T, V, params)  # P not important
        return params["Cv"] * (T - params["T_0"]) * gr

    def volume(self, pressure, temperature, params):
        """
        Returns molar volume. :math:`[m^3]`
        """
        T_0 = params["T_0"]
        V_0 = params["V_0"]
        Cv = params["Cv"]

        a1_ii = 6.0 * params["grueneisen_0"]  # EQ 47
        a2_iikk = (
            -12.0 * params["grueneisen_0"]
            + 36.0 * pow(params["grueneisen_0"], 2.0)
            - 18.0 * params["q_0"] * params["grueneisen_0"]
        )  # EQ 47

        b_iikk = 9.0 * params["K_0"]  # EQ 28
        b_iikkmm = 27.0 * params["K_0"] * (params["Kprime_0"] - 4.0)  # EQ 29z

        # we need to have a sign change in [a,b] to find a zero. Let us start with a
        # conservative guess:
        args = (pressure, temperature, V_0, T_0, Cv, a1_ii, a2_iikk, b_iikk, b_iikkmm)
        try:
            sol = bracket(_delta_pressure, params["V_0"], 1.0e-2 * params["V_0"], args)
        except ValueError:
            raise Exception(
                "Cannot find a volume, perhaps you are outside of the range of validity for the equation of state?"
            )
        return opt.brentq(_delta_pressure, sol[0], sol[1], args=args)

    def pressure(self, temperature, volume, params):
        """
        Returns the pressure of the mineral at a given temperature and volume [Pa]
        """
        gr = self.grueneisen_parameter(
            0.0, temperature, volume, params
        )  # does not depend on pressure

        b_iikk = 9.0 * params["K_0"]  # EQ 28
        b_iikkmm = 27.0 * params["K_0"] * (params["Kprime_0"] - 4.0)  # EQ 29
        f = 0.5 * (pow(params["V_0"] / volume, 2.0 / 3.0) - 1.0)  # EQ 24
        P = (1.0 / 3.0) * (pow(1.0 + 2.0 * f, 5.0 / 2.0)) * (
            (b_iikk * f) + (0.5 * b_iikkmm * pow(f, 2.0))
        ) + gr * params["Cv"] * (
            temperature - params["T_0"]
        ) / volume  # EQ 21

        return P

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns grueneisen parameter :math:`[unitless]`
        """
        return _grueneisen_parameter_fast(
            params["V_0"], volume, params["grueneisen_0"], params["q_0"]
        )

    def isothermal_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`[Pa]`
        """

        E_th_diff = params["Cv"] * (temperature - params["T_0"])
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)
        q = self.volume_dependent_q(params["V_0"] / volume, params)
        K = (
            bm.bulk_modulus(volume, params)
            + (gr + 1.0 - q) * (gr / volume) * E_th_diff
            - (pow(gr, 2.0) / volume) * E_th_diff
        )

        return K

    def isentropic_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus. :math:`[Pa]`
        """
        K_T = self.isothermal_bulk_modulus_reuss(pressure, temperature, volume, params)
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)
        K_S = K_T * (1.0 + gr * alpha * temperature)
        return K_S

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus. :math:`[Pa]`
        """
        T_0 = params["T_0"]
        eta_s = self._isotropic_eta_s(params["V_0"] / volume, params)

        E_th_diff = params["Cv"] * (temperature - params["T_0"])
        return (
            bm.shear_modulus_third_order(volume, params) - eta_s * (E_th_diff) / volume
        )

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant volume. :math:`[J/K/mol]`
        """
        return params["Cv"]

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant pressure. :math:`[J/K/mol]`
        """
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)
        C_v = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        C_p = C_v * (1.0 + gr * alpha * temperature)
        return C_p

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns thermal expansivity. :math:`[1/K]`
        """
        C_v = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)
        K = self.isothermal_bulk_modulus_reuss(pressure, temperature, volume, params)
        alpha = gr * C_v / K / volume
        return alpha

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy at the pressure and temperature of the mineral [J/mol]
        """
        G = (
            self.helmholtz_free_energy(pressure, temperature, volume, params)
            + pressure * volume
        )
        return G

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy at the pressure and temperature of the mineral [J/mol]
        """
        return self.helmholtz_free_energy(
            pressure, temperature, volume, params
        ) + temperature * self.entropy(pressure, temperature, volume, params)

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the entropy at the pressure and temperature of the mineral [J/K/mol]
        """
        S_0 = params["S_0"]
        gruen_0 = params["grueneisen_0"]
        q_0 = params["q_0"]
        S_th = params["Cv"] * (
            np.log(temperature / params["T_0"])
            + _intgroverVdV(params["V_0"], volume, gruen_0, q_0)
        )

        return S_0 + S_th

    def enthalpy(self, pressure, temperature, volume, params):
        """
        Returns the enthalpy at the pressure and temperature of the mineral [J/mol]
        """

        return (
            self.helmholtz_free_energy(pressure, temperature, volume, params)
            + temperature * self.entropy(pressure, temperature, volume, params)
            + pressure * volume
        )

    def helmholtz_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Helmholtz free energy at the pressure and temperature of the mineral [J/mol]
        """
        V_0 = params["V_0"]
        gruen_0 = params["grueneisen_0"]
        q_0 = params["q_0"]
        x = V_0 / volume
        f = 1.0 / 2.0 * (pow(x, 2.0 / 3.0) - 1.0)
        b_iikk = 9.0 * params["K_0"]  # EQ 28
        b_iikkmm = 27.0 * params["K_0"] * (params["Kprime_0"] - 4.0)  # EQ 29

        T_0 = params["T_0"]
        T = temperature
        S_0 = params["S_0"]
        Cv = params["Cv"]
        F_0 = params["E_0"] - T_0 * S_0
        F_cmp = 0.5 * b_iikk * f * f * V_0 + (1.0 / 6.0) * V_0 * b_iikkmm * f * f * f
        F_th = (
            -S_0 * (T - T_0)
            - Cv * (T * np.log(T / T_0) - (T - T_0))
            - Cv * (T - T_0) * _intgroverVdV(V_0, volume, gruen_0, q_0)
        )

        return F_0 + F_cmp + F_th

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """
        if "T_0" not in params:
            params["T_0"] = 300.0

        # If eta_s_0 is not included this is presumably deliberate,
        # as we can model density and bulk modulus just fine without it,
        # so just add it to the dictionary as nan
        # The same goes for the standard state Helmholtz free energy
        if "eta_s_0" not in params:
            params["eta_s_0"] = float("nan")
        if "E_0" not in params:
            params["E_0"] = float("nan")

        # First, let's check the EoS parameters for Tref
        bm.BirchMurnaghanBase.validate_parameters(bm.BirchMurnaghanBase(), params)

        # Now check all the required keys for the
        # thermal part of the EoS are in the dictionary
        expected_keys = ["Cv", "grueneisen_0", "q_0", "eta_s_0"]
        for k in expected_keys:
            if k not in params:
                raise KeyError("params object missing parameter : " + k)

        # Finally, check that the values are reasonable.
        if params["T_0"] < 0.0:
            warnings.warn("Unusual value for T_0", stacklevel=2)
        if params["Cv"] < 0.0 or params["Cv"] > 1000.0:
            warnings.warn("Unusual value for Cv", stacklevel=2)
        if params["grueneisen_0"] < -0.005 or params["grueneisen_0"] > 10.0:
            warnings.warn("Unusual value for grueneisen_0", stacklevel=2)
        if params["q_0"] < -10.0 or params["q_0"] > 10.0:
            warnings.warn("Unusual value for q_0", stacklevel=2)
        if params["eta_s_0"] < -10.0 or params["eta_s_0"] > 10.0:
            warnings.warn("Unusual value for eta_s_0", stacklevel=2)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2022 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import importlib
import numpy as np
from ..utils.chemistry import process_solution_chemistry
from .. import constants
import warnings
from .material import material_property, cached_property
from dataclasses import make_dataclass

Interaction = make_dataclass(
    "Interaction", ["inds", "expts", "f_r", "m_jr", "f_rs", "m_jrs"]
)

try:
    ag = importlib.import_module("autograd")
except ImportError as err:
    print(
        f"Warning: {err}. "
        "For full functionality of BurnMan, please install autograd."
    )


def _ideal_activities_fct(
    molar_fractions,
    endmember_noccupancies,
    n_endmembers,
    n_occupancies,
    site_multiplicities,
    endmember_configurational_entropies,
):
    site_noccupancies = np.dot(molar_fractions, endmember_noccupancies)
    site_multiplicities = np.einsum("i, ij", molar_fractions, site_multiplicities)
    site_occupancies = site_noccupancies * inverseish(site_multiplicities)

    a = np.power(site_occupancies, endmember_noccupancies).prod(-1)
    normalisation_constants = np.exp(
        endmember_configurational_entropies / constants.gas_constant
    )
    return normalisation_constants * a


def _non_ideal_hessian_fct(phi, molar_fractions, n_endmembers, alpha, W):
    q = np.eye(n_endmembers) - phi * np.ones((n_endmembers, n_endmembers))
    sum_pa = np.dot(molar_fractions, alpha)
    hess = np.einsum("m, i, ij, jk, mk->im", -alpha / sum_pa, -alpha, q, W, q)
    hess += hess.T
    return hess


def _non_ideal_interactions_fct(phi, molar_fractions, n_endmembers, alpha, W):
    # -sum(sum(qi.qj.Wij*)
    # equation (2) of Holland and Powell 2003
    q = np.eye(n_endmembers) - phi * np.ones((n_endmembers, n_endmembers))
    # The following are equivalent to
    # np.einsum('i, ij, jk, ik->i', -self.alphas, q, self.Wx, q)
    Wint = -alpha * (q.dot(W) * q).sum(-1)
    return Wint


def _non_ideal_hessian_subreg(p, n_endmembers, Wijk):
    Id = np.identity(n_endmembers)
    IIp = np.einsum("il, jm, k->ijklm", Id, Id, p)
    Ipp = np.einsum("il, j, k->ijkl", Id, p, p)
    ppp = np.einsum("i, j, k->ijk", p, p, p)

    A = (
        IIp
        + np.transpose(IIp, axes=[0, 2, 1, 3, 4])
        + np.transpose(IIp, axes=[1, 0, 2, 3, 4])
        + np.transpose(IIp, axes=[1, 2, 0, 3, 4])
        + np.transpose(IIp, axes=[2, 1, 0, 3, 4])
        + np.transpose(IIp, axes=[2, 0, 1, 3, 4])
    )
    B = 2.0 * (
        Ipp
        + np.transpose(Ipp, axes=[1, 0, 2, 3])
        + np.transpose(Ipp, axes=[2, 1, 0, 3])
    )

    Asum = (
        A - B[:, :, :, :, None] - B[:, :, :, None, :] + 6.0 * ppp[:, :, :, None, None]
    )
    hess = np.einsum("ijklm, ijk->lm", Asum, Wijk)
    return hess


def _non_ideal_interactions_subreg(p, n_endmembers, Wijk):
    Aijkl = np.einsum("li, j, k->ijkl", np.identity(n_endmembers), p, p)
    ppp = np.einsum("i, j, k->ijk", p, p, p)

    Asum = (
        Aijkl
        + np.transpose(Aijkl, axes=[1, 0, 2, 3])
        + np.transpose(Aijkl, axes=[1, 2, 0, 3])
        - 2 * ppp[:, :, :, None]
    )

    Wint = np.einsum("ijk, ijkl->l", Wijk, Asum)
    return Wint


def logish(x, eps=1.0e-7):
    """
    2nd order series expansion of log(x) about eps:
    log(eps) - sum_k=1^infty (f_eps)^k / k
    Prevents infinities at x=0
    """
    f_eps = 1.0 - x / eps
    mask = x > eps
    ln = np.where(x <= eps, np.log(eps) - f_eps - f_eps * f_eps / 2.0, 0.0)
    ln[mask] = np.log(x[mask])
    return ln


def inverseish(x, eps=1.0e-5):
    """
    1st order series expansion of 1/x about eps: 2/eps - x/eps/eps
    Prevents infinities at x=0
    """
    mask = x > eps
    oneoverx = np.where(x <= eps, 2.0 / eps - x / eps / eps, 0.0)
    oneoverx[mask] = 1.0 / x[mask]
    return oneoverx


def dpdn(molar_amounts, n, ones, eye):
    """
    The partial derivative of endmember proportions
    with respect to endmember amounts.

    :param molar_amounts: molar amounts of independent endmembers
    :type molar_fractions: 1D numpy array
    :param n: sum of the endmember amounts (usually equal to one)
    :type n: float
    :param ones: a vector of ones of length equal to the number of endmembers
    :type ones: 1D numpy array
    :param eye: the identity matrix of size equal to the number of endmembers
    :type eye: 2D numpy array
    """
    return (eye - np.einsum("k, m->km", molar_amounts, ones)) / n


def d2pdndn(molar_amounts, nsqr, ones, eyeones):
    """
    The second partial derivative of endmember proportions
    with respect to endmember amounts.

    :param molar_amounts: molar amounts of independent endmembers
    :type molar_fractions: 1D numpy array
    :param nsqr: square of the sum of the endmember amounts (usually equal to one)
    :type n: float
    :param ones: a vector of ones of length equal to the number of endmembers
    :type ones: 1D numpy array
    :param eyeones: delta_ij 1_k + delta_ik 1_j
    :type eyeones: 3D numpy array
    """

    return (2.0 * np.einsum("n, m, k->knm", ones, ones, molar_amounts) - eyeones) / nsqr


class SolutionModel(object):
    """
    This is the base class for a solution model,  intended for use
    in defining solutions and performing thermodynamic calculations
    on them.  All minerals of type :class:`burnman.Solution` use
    a solution model for defining how the endmembers in the solution
    interact.

    A user wanting a new solution model should define the functions included
    in the base class. All of the functions in the base class return zero,
    so if the user-defined solution model does not implement them,
    they essentially have no effect, and the Gibbs free energy and molar
    volume of a solution will be equal to the weighted arithmetic
    averages of the different endmember values.
    """

    def __init__(self):
        """
        Does nothing.
        """
        pass

    def excess_gibbs_free_energy(self, pressure, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess Gibbs free energy of the solution.
        The base class implementation assumes that the excess gibbs
        free energy is zero.

        :param pressure: Pressure at which to evaluate the solution model [Pa].
        :type pressure: float
        :param temperature: Temperature at which to evaluate the solution model [K].
        :type temperature: float
        :param molar_fractions: List of molar fractions of the
            different independent endmembers in the solution model.
        :type molar_fractions: list of floats

        :returns: The excess Gibbs energy.
        :rtype: float
        """
        return np.dot(
            np.array(molar_fractions),
            self.excess_partial_gibbs_free_energies(
                pressure, temperature, molar_fractions
            ),
        )

    def excess_volume(self, pressure, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess volume of the solution.
        The base class implementation assumes that the excess volume is zero.

        :param pressure: Pressure at which to evaluate the solution model [Pa].
        :type pressure: float
        :param temperature: Temperature at which to evaluate the solution model [K].
        :type temperature: float
        :param molar_fractions: List of molar fractions of the
            different independent endmembers in the solution model.
        :type molar_fractions: list of floats

        :returns: The excess volume of the solution.
        :rtype: float
        """
        return np.dot(
            molar_fractions,
            self.excess_partial_volumes(pressure, temperature, molar_fractions),
        )

    def excess_entropy(self, pressure, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess entropy of the solution.
        The base class implementation assumes that the excess entropy is zero.

        :param pressure: Pressure at which to evaluate the solution model [Pa].
        :type pressure: float
        :param temperature: Temperature at which to evaluate the solution model [K].
        :type temperature: float
        :param molar_fractions: List of molar fractions of the
            different independent endmembers in the solution model.
        :type molar_fractions: list of floats

        :returns: The excess entropy of the solution.
        :rtype: float
        """
        return np.dot(
            molar_fractions,
            self.excess_partial_entropies(pressure, temperature, molar_fractions),
        )

    def excess_enthalpy(self, pressure, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess enthalpy of the solution.
        The base class implementation assumes that the excess enthalpy is zero.

        :param pressure: Pressure at which to evaluate the solution model [Pa].
        :type pressure: float
        :param temperature: Temperature at which to evaluate the solution model [K].
        :type temperature: float
        :param molar_fractions: List of molar fractions of the
            different independent endmembers in the solution model.
        :type molar_fractions: list of floats

        :returns: The excess enthalpy of the solution.
        :rtype: float
        """
        return self.excess_gibbs_free_energy(
            pressure, temperature, molar_fractions
        ) + temperature * self.excess_entropy(pressure, temperature, molar_fractions)

    def excess_partial_gibbs_free_energies(
        self, pressure, temperature, molar_fractions
    ):
        """
        Given a list of molar fractions of different phases,
        compute the excess Gibbs free energy for each endmember
        of the solution.
        The base class implementation assumes that the excess gibbs
        free energy is zero.

        :param pressure: Pressure at which to evaluate the solution model [Pa].
        :type pressure: float
        :param temperature: Temperature at which to evaluate the solution model [K].
        :type temperature: float
        :param molar_fractions: List of molar fractions of the
            different independent endmembers in the solution model.
        :type molar_fractions: list of floats

        :returns: The excess partial Gibbs free energy of each endmember.
        :rtype: numpy.array
        """
        return np.zeros_like(molar_fractions)

    def excess_partial_entropies(self, pressure, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess entropy for each endmember of the solution.
        The base class implementation assumes that the excess entropy
        is zero (true for mechanical solutions).

        :param pressure: Pressure at which to evaluate the solution model [Pa].
        :type pressure: float
        :param temperature: Temperature at which to evaluate the solution model [K].
        :type temperature: float
        :param molar_fractions: List of molar fractions of the
            different independent endmembers in the solution model.
        :type molar_fractions: list of floats

        :returns: The excess partial entropy of each endmember.
        :rtype: numpy.array
        """
        return np.zeros_like(molar_fractions)

    def excess_partial_volumes(self, pressure, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess volume for each endmember of the solution.
        The base class implementation assumes that the excess volume
        is zero.

        :param pressure: Pressure at which to evaluate the solution model [Pa].
        :type pressure: float
        :param temperature: Temperature at which to evaluate the solution model [K].
        :type temperature: float
        :param molar_fractions: List of molar fractions of the
            different independent endmembers in the solution model.
        :type molar_fractions: list of floats

        :returns: The excess partial volume of each endmember.
        :rtype: numpy.array
        """
        return np.zeros_like(np.array(molar_fractions))

    def Cp_excess(self):
        """
        Returns the excess heat capacity of the solution model
        at its current state
        """
        return 0.0

    def alphaV_excess(self):
        """
        Returns the excess alpha*V of the solution model
        at its current state
        """
        return 0.0

    def VoverKT_excess(self):
        """
        Returns the excess V/K_T of the solution model
        at its current state
        """
        return 0.0


class MechanicalSolution(SolutionModel):
    """
    An extremely simple class representing a mechanical solution model.
    A mechanical solution experiences no interaction between endmembers.
    Therefore, unlike ideal solutions there is no entropy of mixing;
    the total gibbs free energy of the solution is equal to the
    dot product of the molar gibbs free energies and molar fractions
    of the constituent materials.
    """

    def __init__(self, endmembers):
        self.endmembers = endmembers
        self.n_endmembers = len(endmembers)
        self.formulas = [e[1] for e in endmembers]

    def excess_gibbs_free_energy(self, pressure, temperature, molar_fractions):
        return 0.0

    def excess_volume(self, pressure, temperature, molar_fractions):
        return 0.0

    def excess_entropy(self, pressure, temperature, molar_fractions):
        return 0.0

    def excess_enthalpy(self, pressure, temperature, molar_fractions):
        return 0.0

    def excess_partial_gibbs_free_energies(
        self, pressure, temperature, molar_fractions
    ):
        return np.zeros_like(molar_fractions)

    def excess_partial_volumes(self, pressure, temperature, molar_fractions):
        return np.zeros_like(molar_fractions)

    def excess_partial_entropies(self, pressure, temperature, molar_fractions):
        return np.zeros_like(molar_fractions)

    def activity_coefficients(self, pressure, temperature, molar_fractions):
        return np.ones_like(molar_fractions)

    def activities(self, pressure, temperature, molar_fractions):
        return np.ones_like(molar_fractions)


class IdealSolution(SolutionModel):
    """
    A class representing an ideal solution model.
    Calculates the excess gibbs free energy and entropy due to configurational
    entropy. Excess internal energy and volume are equal to zero.

    The multiplicity of each type of site in the structure is allowed to
    change linearly as a function of endmember proportions. This class
    is therefore equivalent to the entropic part of
    a Temkin-type model :cite:`Temkin1945`.
    """

    def __init__(self, endmembers):
        self.endmembers = endmembers
        self.n_endmembers = len(endmembers)
        self.formulas = [e[1] for e in endmembers]

        # Process solution chemistry
        process_solution_chemistry(self)

        self._calculate_endmember_configurational_entropies()

    def _calculate_endmember_configurational_entropies(self):
        S_conf = -(
            constants.gas_constant
            * (
                self.endmember_noccupancies
                * (
                    logish(self.endmember_noccupancies)
                    - logish(self.site_multiplicities)
                )
            ).sum(-1)
        )
        self.endmember_configurational_entropies = S_conf

    def excess_partial_gibbs_free_energies(
        self, pressure, temperature, molar_fractions
    ):
        return self._ideal_excess_partial_gibbs(temperature, molar_fractions)

    def excess_partial_entropies(self, pressure, temperature, molar_fractions):
        return self._ideal_excess_partial_entropies(temperature, molar_fractions)

    def excess_partial_volumes(self, pressure, temperature, molar_fractions):
        return np.zeros((self.n_endmembers))

    def gibbs_hessian(self, pressure, temperature, molar_fractions):
        hess_S = self._ideal_entropy_hessian(temperature, molar_fractions)
        return -temperature * hess_S

    def entropy_hessian(self, pressure, temperature, molar_fractions):
        hess_S = self._ideal_entropy_hessian(temperature, molar_fractions)
        return hess_S

    def volume_hessian(self, pressure, temperature, molar_fractions):
        return np.zeros((len(molar_fractions), len(molar_fractions)))

    def _configurational_entropy(self, molar_fractions):
        site_noccupancies = np.einsum(
            "i, ij", molar_fractions, self.endmember_noccupancies
        )
        site_multiplicities = np.einsum(
            "i, ij", molar_fractions, self.site_multiplicities
        )
        site_occupancies = site_noccupancies * inverseish(site_multiplicities)
        conf_entropy = -(
            constants.gas_constant
            * (site_noccupancies * logish(site_occupancies)).sum(-1)
        )
        return conf_entropy

    def _ideal_excess_partial_gibbs(self, temperature, molar_fractions):
        return -(
            temperature
            * self._ideal_excess_partial_entropies(temperature, molar_fractions)
        )

    def _ideal_excess_partial_entropies(self, temperature, molar_fractions):
        return -(constants.gas_constant * self._log_ideal_activities(molar_fractions))

    def _ideal_entropy_hessian(self, temperature, molar_fractions):
        hessian = -(
            constants.gas_constant
            * self._log_ideal_activity_derivatives(molar_fractions)
        )
        return hessian

    def _log_ideal_activities(self, molar_fractions):
        site_noccupancies = np.einsum(
            "i, ij", molar_fractions, self.endmember_noccupancies
        )
        site_multiplicities = np.einsum(
            "i, ij", molar_fractions, self.site_multiplicities
        )

        lna = np.einsum(
            "ij, j->i",
            self.endmember_noccupancies,
            logish(site_noccupancies) - logish(site_multiplicities),
        )
        normalisation_constants = (
            self.endmember_configurational_entropies / constants.gas_constant
        )
        return lna + normalisation_constants

    def _log_ideal_activity_derivatives(self, molar_fractions):
        site_noccupancies = np.einsum(
            "i, ij", molar_fractions, self.endmember_noccupancies
        )
        site_multiplicities = np.einsum(
            "i, ij", molar_fractions, self.site_multiplicities
        )

        dlnadp = np.einsum(
            "pj, qj, j->pq",
            self.endmember_noccupancies,
            self.endmember_noccupancies,
            inverseish(site_noccupancies),
        ) - np.einsum(
            "pj, qj, j->pq",
            self.endmember_noccupancies,
            self.site_multiplicities,
            inverseish(site_multiplicities),
        )

        return dlnadp

    def _ideal_activities(self, molar_fractions):
        return _ideal_activities_fct(
            molar_fractions,
            self.endmember_noccupancies,
            self.n_endmembers,
            self.n_occupancies,
            self.site_multiplicities,
            self.endmember_configurational_entropies,
        )

    def activity_coefficients(self, pressure, temperature, molar_fractions):
        return np.ones_like(molar_fractions)

    def activities(self, pressure, temperature, molar_fractions):
        return self._ideal_activities(molar_fractions)

    @cached_property
    def ones(self):
        """
        A vector of ones with length equal to the number of endmembers
        :return: ones
        :rtype: 1D numpy array
        """
        return np.ones(self.n_endmembers)

    @cached_property
    def eye(self):
        """
        An identity matrix with size equal to the number of endmembers
        :return: eye
        :rtype: 2D numpy array
        """
        return np.eye(self.n_endmembers)

    @cached_property
    def eyeones(self):
        """
        A convenience function consisting of two concatenations
        of an identity matrix and ones vector with size
        equal to the number of endmembers.
        :return: delta_ij 1_k + delta_ik 1_j
        :rtype: 3D numpy array
        """
        return np.einsum("km, n->kmn", self.eye, self.ones) + np.einsum(
            "kn, m->kmn", self.eye, self.ones
        )


class AsymmetricRegularSolution(IdealSolution):
    """
    Solution model implementing the asymmetric regular solution model
    formulation as described in :cite:`HP2003`.

    The excess nonconfigurational Gibbs energy is given by the
    expression:

    .. math::
        \\mathcal{G}_{\\textrm{excess}} = \\alpha^T p (\\phi^T W \\phi)

    :math:`\\alpha` is a vector of van Laar parameters governing asymmetry
    in the excess properties.

    .. math::
        \\phi_i = \\frac{\\alpha_i p_i}{\\sum_{k=1}^{n} \\alpha_k p_k},
        W_{ij} = \\frac{2 w_{ij}}{\\alpha_i + \\alpha_j} \\textrm{for i<j}
    """

    def __init__(
        self,
        endmembers,
        alphas,
        energy_interaction,
        volume_interaction=None,
        entropy_interaction=None,
    ):
        self.n_endmembers = len(endmembers)

        # Create array of van Laar parameters
        self.alphas = np.array(alphas)

        # Create 2D arrays of interaction parameters
        self.We = np.triu(2.0 / (self.alphas[:, np.newaxis] + self.alphas), 1)
        self.We[np.triu_indices(self.n_endmembers, 1)] *= np.array(
            [i for row in energy_interaction for i in row]
        )

        if entropy_interaction is not None:
            self.Ws = np.triu(2.0 / (self.alphas[:, np.newaxis] + self.alphas), 1)
            self.Ws[np.triu_indices(self.n_endmembers, 1)] *= np.array(
                [i for row in entropy_interaction for i in row]
            )
        else:
            self.Ws = np.zeros((self.n_endmembers, self.n_endmembers))

        if volume_interaction is not None:
            self.Wv = np.triu(2.0 / (self.alphas[:, np.newaxis] + self.alphas), 1)
            self.Wv[np.triu_indices(self.n_endmembers, 1)] *= np.array(
                [i for row in volume_interaction for i in row]
            )
        else:
            self.Wv = np.zeros((self.n_endmembers, self.n_endmembers))

        # initialize ideal solution model
        IdealSolution.__init__(self, endmembers)

    def _phi(self, molar_fractions):
        phi = self.alphas * molar_fractions
        phi = np.divide(phi, np.sum(phi))
        return phi

    def _non_ideal_interactions(self, W, molar_fractions):
        # -sum(sum(qi.qj.Wij*)
        # equation (2) of Holland and Powell 2003
        phi = self._phi(molar_fractions)
        return _non_ideal_interactions_fct(
            phi, np.array(molar_fractions), self.n_endmembers, self.alphas, W
        )

    def _non_ideal_excess_partial_gibbs(self, pressure, temperature, molar_fractions):
        Eint = self._non_ideal_interactions(self.We, molar_fractions)
        Sint = self._non_ideal_interactions(self.Ws, molar_fractions)
        Vint = self._non_ideal_interactions(self.Wv, molar_fractions)
        return Eint - temperature * Sint + pressure * Vint

    def excess_partial_gibbs_free_energies(
        self, pressure, temperature, molar_fractions
    ):
        ideal_gibbs = IdealSolution._ideal_excess_partial_gibbs(
            self, temperature, molar_fractions
        )
        non_ideal_gibbs = self._non_ideal_excess_partial_gibbs(
            pressure, temperature, molar_fractions
        )
        return ideal_gibbs + non_ideal_gibbs

    def excess_partial_entropies(self, pressure, temperature, molar_fractions):
        ideal_entropies = IdealSolution._ideal_excess_partial_entropies(
            self, temperature, molar_fractions
        )
        non_ideal_entropies = self._non_ideal_interactions(self.Ws, molar_fractions)
        return ideal_entropies + non_ideal_entropies

    def excess_partial_volumes(self, pressure, temperature, molar_fractions):
        return self._non_ideal_interactions(self.Wv, molar_fractions)

    def gibbs_hessian(self, pressure, temperature, molar_fractions):
        ideal_entropy_hessian = IdealSolution._ideal_entropy_hessian(
            self, temperature, molar_fractions
        )
        phi = self._phi(molar_fractions)
        nonideal_gibbs_hessian = _non_ideal_hessian_fct(
            phi,
            molar_fractions,
            self.n_endmembers,
            self.alphas,
            self.We - temperature * self.Ws + pressure * self.Wv,
        )

        return nonideal_gibbs_hessian - temperature * ideal_entropy_hessian

    def entropy_hessian(self, pressure, temperature, molar_fractions):
        ideal_entropy_hessian = IdealSolution._ideal_entropy_hessian(
            self, temperature, molar_fractions
        )
        phi = self._phi(molar_fractions)
        nonideal_entropy_hessian = _non_ideal_hessian_fct(
            phi, molar_fractions, self.n_endmembers, self.alphas, self.Ws
        )
        return ideal_entropy_hessian + nonideal_entropy_hessian

    def volume_hessian(self, pressure, temperature, molar_fractions):
        phi = self._phi(molar_fractions)
        return _non_ideal_hessian_fct(
            phi, molar_fractions, self.n_endmembers, self.alphas, self.Wv
        )

    def activity_coefficients(self, pressure, temperature, molar_fractions):
        if temperature > 1.0e-10:
            return np.exp(
                self._non_ideal_excess_partial_gibbs(
                    pressure, temperature, molar_fractions
                )
                / (constants.gas_constant * temperature)
            )
        else:
            raise Exception("Activity coefficients not defined at 0 K.")

    def activities(self, pressure, temperature, molar_fractions):
        return IdealSolution.activities(
            self, pressure, temperature, molar_fractions
        ) * self.activity_coefficients(pressure, temperature, molar_fractions)


class SymmetricRegularSolution(AsymmetricRegularSolution):
    """
    Solution model implementing the symmetric regular solution model.
    This is a special case of the
    :class:`burnman.solutionmodel.AsymmetricRegularSolution` class.
    """

    def __init__(
        self,
        endmembers,
        energy_interaction,
        volume_interaction=None,
        entropy_interaction=None,
    ):
        alphas = np.ones(len(endmembers))
        AsymmetricRegularSolution.__init__(
            self,
            endmembers,
            alphas,
            energy_interaction,
            volume_interaction,
            entropy_interaction,
        )


class SubregularSolution(IdealSolution):
    """
    Solution model implementing the subregular solution model formulation
    as described in :cite:`HW1989`. The excess nonconfigurational
    Gibbs energy is given by the expression:

    .. math::
        \\mathcal{G}_{\\textrm{excess}} = \\sum_i \\sum_{j > i} (p_i p_j^2
        W_{ij} + p_j p_i^2 W_{ji} + \\sum_{k > j > i} p_i p_j p_k W_{ijk})

    Interaction parameters are inserted into a 3D interaction matrix during
    initialization to make use of numpy vector algebra.

    :param endmembers: A list of all the independent endmembers in the solution.
        The first item of each list gives the Mineral object corresponding
        to the endmember. The second item gives the site-species formula.
    :type endmembers: list of lists
    :param energy_interaction: The binary endmember interaction energies.
        Each interaction[i, j-i-1, 0] corresponds to W(i,j), while
        interaction[i, j-i-1, 1] corresponds to W(j,i).
    :type energy_interaction: list of list of lists
    :param volume_interaction: The binary endmember interaction volumes.
        Each interaction[i, j-i-1, 0] corresponds to W(i,j), while
        interaction[i, j-i-1, 1] corresponds to W(j,i).
    :type volume_interaction: list of list of lists
    :param entropy_interaction: The binary endmember interaction entropies.
        Each interaction[i, j-i-1, 0] corresponds to W(i,j), while
        interaction[i, j-i-1, 1] corresponds to W(j,i).
    :type entropy_interaction: list of list of lists
    :param energy_ternary_terms: The ternary interaction energies.
        Each list should contain four entries:
        the indices i, j, k and the value of the interaction.
    :type energy_ternary_terms: list of lists
    :param volume_ternary_terms: The ternary interaction volumes.
        Each list should contain four entries:
        the indices i, j, k and the value of the interaction.
    :type volume_ternary_terms: list of lists
    :param entropy_ternary_terms: The ternary interaction entropies.
        Each list should contain four entries:
        the indices i, j, k and the value of the interaction.
    :type entropy_ternary_terms: list of lists
    """

    def __init__(
        self,
        endmembers,
        energy_interaction,
        volume_interaction=None,
        entropy_interaction=None,
        energy_ternary_terms=None,
        volume_ternary_terms=None,
        entropy_ternary_terms=None,
    ):
        """
        Initialization function for the SubregularSolution class.
        """

        self.n_endmembers = len(endmembers)

        # Create 3D arrays of interaction parameters
        self.Wijke = np.zeros(
            shape=(self.n_endmembers, self.n_endmembers, self.n_endmembers)
        )
        self.Wijks = np.zeros_like(self.Wijke)
        self.Wijkv = np.zeros_like(self.Wijke)

        # setup excess enthalpy interaction matrix
        for i in range(self.n_endmembers):
            for j in range(i + 1, self.n_endmembers):
                w0 = energy_interaction[i][j - i - 1][0] / 2.0
                w1 = energy_interaction[i][j - i - 1][1] / 2.0
                self.Wijke[:, i, j] += w0
                self.Wijke[:, j, i] += w1

                self.Wijke[i, j, j] += w0
                self.Wijke[j, i, i] += w1

                self.Wijke[i, j, i] -= w0
                self.Wijke[j, i, j] -= w1

        if energy_ternary_terms is not None:
            for i, j, k, v in energy_ternary_terms:
                self.Wijke[i, j, k] += v

        if entropy_interaction is not None:
            for i in range(self.n_endmembers):
                for j in range(i + 1, self.n_endmembers):
                    w0 = entropy_interaction[i][j - i - 1][0] / 2.0
                    w1 = entropy_interaction[i][j - i - 1][1] / 2.0
                    self.Wijks[:, i, j] += w0
                    self.Wijks[:, j, i] += w1

                    self.Wijks[i, j, j] += w0
                    self.Wijks[j, i, i] += w1

                    self.Wijks[i, j, i] -= w0
                    self.Wijks[j, i, j] -= w1

        if entropy_ternary_terms is not None:
            for i, j, k, v in entropy_ternary_terms:
                self.Wijks[i, j, k] += v

        if volume_interaction is not None:
            for i in range(self.n_endmembers):
                for j in range(i + 1, self.n_endmembers):
                    w0 = volume_interaction[i][j - i - 1][0] / 2.0
                    w1 = volume_interaction[i][j - i - 1][1] / 2.0
                    self.Wijkv[:, i, j] += w0
                    self.Wijkv[:, j, i] += w1

                    self.Wijkv[i, j, j] += w0
                    self.Wijkv[j, i, i] += w1

                    self.Wijkv[i, j, i] -= w0
                    self.Wijkv[j, i, j] -= w1

        if volume_ternary_terms is not None:
            for i, j, k, v in volume_ternary_terms:
                self.Wijkv[i, j, k] += v

        # initialize ideal solution model
        IdealSolution.__init__(self, endmembers)

    def _non_ideal_function(self, Wijk, molar_fractions):
        n = len(molar_fractions)
        return _non_ideal_interactions_subreg(molar_fractions, n, Wijk)

    def _non_ideal_interactions(self, molar_fractions):
        # equation (6') of Helffrich and Wood, 1989
        Eint = self._non_ideal_function(self.Wijke, molar_fractions)
        Sint = self._non_ideal_function(self.Wijks, molar_fractions)
        Vint = self._non_ideal_function(self.Wijkv, molar_fractions)
        return Eint, Sint, Vint

    def _non_ideal_excess_partial_gibbs(self, pressure, temperature, molar_fractions):
        Eint, Sint, Vint = self._non_ideal_interactions(molar_fractions)
        return Eint - temperature * Sint + pressure * Vint

    def excess_partial_gibbs_free_energies(
        self, pressure, temperature, molar_fractions
    ):
        ideal_gibbs = IdealSolution._ideal_excess_partial_gibbs(
            self, temperature, molar_fractions
        )
        non_ideal_gibbs = self._non_ideal_excess_partial_gibbs(
            pressure, temperature, molar_fractions
        )
        return ideal_gibbs + non_ideal_gibbs

    def excess_partial_entropies(self, pressure, temperature, molar_fractions):
        ideal_entropies = IdealSolution._ideal_excess_partial_entropies(
            self, temperature, molar_fractions
        )
        non_ideal_entropies = self._non_ideal_function(self.Wijks, molar_fractions)
        return ideal_entropies + non_ideal_entropies

    def excess_partial_volumes(self, pressure, temperature, molar_fractions):
        non_ideal_volumes = self._non_ideal_function(self.Wijkv, molar_fractions)
        return non_ideal_volumes

    def gibbs_hessian(self, pressure, temperature, molar_fractions):
        n = len(molar_fractions)
        ideal_entropy_hessian = IdealSolution._ideal_entropy_hessian(
            self, temperature, molar_fractions
        )
        nonideal_gibbs_hessian = _non_ideal_hessian_subreg(
            molar_fractions,
            n,
            self.Wijke - temperature * self.Wijks + pressure * self.Wijkv,
        )

        return nonideal_gibbs_hessian - temperature * ideal_entropy_hessian

    def entropy_hessian(self, pressure, temperature, molar_fractions):
        n = len(molar_fractions)
        ideal_entropy_hessian = IdealSolution._ideal_entropy_hessian(
            self, temperature, molar_fractions
        )
        nonideal_entropy_hessian = _non_ideal_hessian_subreg(
            molar_fractions, n, self.Wijks
        )
        return ideal_entropy_hessian + nonideal_entropy_hessian

    def volume_hessian(self, pressure, temperature, molar_fractions):
        n = len(molar_fractions)
        return _non_ideal_hessian_subreg(molar_fractions, n, self.Wijkv)

    def activity_coefficients(self, pressure, temperature, molar_fractions):
        if temperature > 1.0e-10:
            return np.exp(
                self._non_ideal_excess_partial_gibbs(
                    pressure, temperature, molar_fractions
                )
                / (constants.gas_constant * temperature)
            )
        else:
            raise Exception("Activity coefficients not defined at 0 K.")

    def activities(self, pressure, temperature, molar_fractions):
        return IdealSolution.activities(
            self, pressure, temperature, molar_fractions
        ) * self.activity_coefficients(pressure, temperature, molar_fractions)


class FunctionSolution(IdealSolution):
    """
    Solution model implementing a generalized solution model.
    The extensive excess nonconfigurational Gibbs energy is
    provided as a function by the user.

    Derivatives are calculated using the autograd module,
    and so the user-defined excess Gibbs energy function
    should be defined using autograd-friendly expressions.

    :param endmembers: A list of all the independent endmembers in the solution.
        The first item of each list gives the Mineral object corresponding
        to the endmember. The second item gives the site-species formula.
    :type endmembers: list of lists

    :param excess_gibbs_function: The nonconfigurational Gibbs energy function
        with arguments pressure, temperature and molar_amounts, in that order.
        Note that the function must be extensive; if the molar amounts
        are doubled, the Gibbs energy must also double.
    :type excess_gibbs_function: function
    """

    def __init__(self, endmembers, excess_gibbs_function):
        """
        Initialization function for the GeneralSolution class.
        """

        # initialize ideal solution model
        IdealSolution.__init__(self, endmembers)

        self.n_endmembers = len(endmembers)
        self._excess_gibbs_function = excess_gibbs_function

        self._non_ideal_excess_partial_gibbs = ag.jacobian(
            excess_gibbs_function, argnum=2
        )

        def partial_entropies(pressure, temperature, molar_amounts):
            with warnings.catch_warnings(record=True):
                warnings.simplefilter("always")
                return -ag.jacobian(self._non_ideal_excess_partial_gibbs, argnum=1)(
                    pressure, temperature, molar_amounts
                )

        self._non_ideal_excess_partial_entropies = partial_entropies

        def partial_volumes(pressure, temperature, molar_amounts):
            with warnings.catch_warnings(record=True):
                warnings.simplefilter("always")
                return ag.jacobian(self._non_ideal_excess_partial_gibbs, argnum=0)(
                    pressure, temperature, molar_amounts
                )

        self.excess_partial_volumes = partial_volumes

        self._non_ideal_gibbs_hessian = ag.jacobian(
            self._non_ideal_excess_partial_gibbs, argnum=2
        )

        def entropy_hess(pressure, temperature, molar_amounts):
            with warnings.catch_warnings(record=True):
                warnings.simplefilter("always")
                return ag.jacobian(partial_entropies, argnum=2)(
                    pressure, temperature, molar_amounts
                )

        self._non_ideal_entropy_hessian = entropy_hess

        def volume_hess(pressure, temperature, molar_amounts):
            with warnings.catch_warnings(record=True):
                warnings.simplefilter("always")
                return ag.jacobian(partial_volumes, argnum=2)(
                    pressure, temperature, molar_amounts
                )

        self.volume_hessian = volume_hess

    def excess_partial_gibbs_free_energies(
        self, pressure, temperature, molar_fractions
    ):
        ideal_gibbs = IdealSolution._ideal_excess_partial_gibbs(
            self, temperature, molar_fractions
        )
        non_ideal_gibbs = self._non_ideal_excess_partial_gibbs(
            pressure, temperature, molar_fractions
        )
        return ideal_gibbs + non_ideal_gibbs

    def excess_partial_entropies(self, pressure, temperature, molar_fractions):
        ideal_entropies = IdealSolution._ideal_excess_partial_entropies(
            self, temperature, molar_fractions
        )
        non_ideal_entropies = self._non_ideal_excess_partial_entropies(
            pressure, temperature, molar_fractions
        )
        return ideal_entropies + non_ideal_entropies

    def gibbs_hessian(self, pressure, temperature, molar_fractions):
        ideal_entropy_hessian = IdealSolution._ideal_entropy_hessian(
            self, temperature, molar_fractions
        )
        nonideal_gibbs_hessian = self._non_ideal_gibbs_hessian(
            pressure, temperature, molar_fractions
        )

        return nonideal_gibbs_hessian - temperature * ideal_entropy_hessian

    def entropy_hessian(self, pressure, temperature, molar_fractions):
        ideal_entropy_hessian = IdealSolution._ideal_entropy_hessian(
            self, temperature, molar_fractions
        )
        nonideal_entropy_hessian = self._non_ideal_entropy_hessian(
            pressure, temperature, molar_fractions
        )
        return ideal_entropy_hessian + nonideal_entropy_hessian

    def activity_coefficients(self, pressure, temperature, molar_fractions):
        if temperature > 1.0e-10:
            return np.exp(
                self._non_ideal_excess_partial_gibbs(
                    pressure, temperature, molar_fractions
                )
                / (constants.gas_constant * temperature)
            )
        else:
            raise Exception("Activity coefficients not defined at 0 K.")

    def activities(self, pressure, temperature, molar_fractions):
        return IdealSolution.activities(
            self, pressure, temperature, molar_fractions
        ) * self.activity_coefficients(pressure, temperature, molar_fractions)


class PolynomialSolution(IdealSolution):
    """
    Solution model implementing a general polynomial solution model.

    :param endmembers: A list of all the independent endmembers in the solution.
        The first item of each list gives the Mineral object corresponding
        to the endmember. The second item gives the site-species formula.
    :type endmembers: list of lists
    :param ESV_interactions: A list containing lists where the first three elements are
        energy, entropy and volume interactions and the rest of the elements
        are indices of the transformed endmembers to which those
        interactions correspond, and the proportion exponents.
        For example, [2., 0., 0., 0, 4, 1, 1] would correspond to an interaction
        of 2*p'[0]^4*p'[1]^1.
    :type ESV_interactions: list of lists
    :param interaction_endmembers: A list of minerals involved in the interaction terms.
    :type interaction_endmembers: list of :class:`burnman.Mineral` objects
    :param endmember_coefficients_and_interactions: list of lists
        A list containing lists where the first n elements are
        coefficients for each of the interaction_endmembers and the
        rest of the elements are indices of the transformed
        endmembers to which those interactions correspond.
        For example, [1., 0., -1., 0, 4, 1, 1] would correspond to an interaction
        of (mbr[0].gibbs - mbr[2].gibbs)*p'[0]^4*p'[1]^1.
    :type endmember_coefficients_and_interactions: list of lists
    :param transformation_matrix: The interactions for a given solution may
        be most compactly expressed not as a polynomial function of the
        proportions of the endmembers, but a polynomial function of a
        linearly transformed set. This parameter is a square numpy array A,
        where p'i = A_ij p_j
    :type transformation_matrix: 2D numpy array
    """

    def __init__(
        self,
        endmembers,
        ESV_interactions=None,
        interaction_endmembers=[],
        endmember_coefficients_and_interactions=None,
        transformation_matrix=None,
    ):
        # initialize ideal solution model
        IdealSolution.__init__(self, endmembers)

        self.n_endmembers = len(endmembers)
        self.endmembers = endmembers

        self.W_ESV = None
        self.c_ESV = None
        self.n_W_ESV = 0
        if ESV_interactions is not None:
            self.W_ESV, self.c_ESV = self._make_interaction_arrays(ESV_interactions, 3)
            self.n_W_ESV = len(ESV_interactions)

        self.n_interaction_endmembers = len(interaction_endmembers)
        self.interaction_endmembers = interaction_endmembers

        self.W_mbr = None
        self.c_mbr = None
        self.n_W_mbr = 0
        if self.n_interaction_endmembers > 0:
            self.W_mbr, self.c_mbr = self._make_interaction_arrays(
                endmember_coefficients_and_interactions, self.n_interaction_endmembers
            )
            self.n_W_mbr = len(endmember_coefficients_and_interactions)

        if transformation_matrix is None:
            self.dqdp = np.eye(self.n_endmembers)
        else:
            self.dqdp = transformation_matrix
        self.n_transformed_endmembers = len(self.dqdp)
        self.reset()

    def reset(self):
        """
        Resets all cached material properties.
        It is typically not required for the user to call this function.
        """
        self._cached = {}

    def set_composition(self, molar_fractions):
        """
        Resets all cached material properties.
        It is typically not required for the user to call this function.
        """
        self.reset()
        self.molar_fractions = molar_fractions
        self.trans_fractions = np.einsum("ij, j->i", self.dqdp, molar_fractions)

    def set_state(self, pressure, temperature):
        """
        Sets the states for the interaction endmembers.
        It is typically not required for the user to call this function.
        """
        for mbr in self.interaction_endmembers:
            mbr.set_state(pressure, temperature)

    def _make_prefactors_and_exponents(self, a):
        """
        This hidden method makes an Interaction object
        containing convenience arrays to facilitate rapid
        calculation of the first and second compositional
        derivatives of the Gibbs, entropy and volume
        excesses.

        :param a: A list of polynomial indices and exponents.
        :type a: list
        :return: An interaction object containing the
            indices (inds), exponents (expts),
            first derivative prefactors (j_r) and exponents (m_jr) and
            second derivative prefactors (f_rs) and exponents (m_jrs).
        :rtype: Interaction dataclass instance
        """
        # assign exponent matrix for first derivatives
        if len(a) % 2 != 0:
            raise Exception(
                "Interaction must have paired indices " f"and exponents, currently {a}"
            )
        indices = a[::2]
        exponents = np.array(a[1::2])
        n_indices = len(indices)
        m_jr = np.zeros((n_indices, self.n_endmembers))
        m_jr[:, :] = exponents[:, np.newaxis]
        m_jr[list(range(len(m_jr))), indices] -= np.ones(len(indices))

        # assign factors for first derivatives
        f_jr = np.zeros((n_indices, self.n_endmembers))
        f_jr[:, indices] = 1.0
        f_jr[range(len(m_jr)), indices] = exponents
        f_r = np.prod(f_jr, axis=0)

        # assign exponent matrix for second derivatives
        m_jrs = np.zeros((n_indices, self.n_endmembers, self.n_endmembers))
        m_jrs[:, :, :] = exponents[:, np.newaxis, np.newaxis]

        m_jrs -= (
            np.einsum(
                "ij, k->ijk", np.eye(self.n_endmembers), np.ones(self.n_endmembers)
            )
            + np.einsum(
                "ik, j->ijk", np.eye(self.n_endmembers), np.ones(self.n_endmembers)
            )
        )[indices]

        # assign factors for second derivatives
        f_jrs = np.zeros((n_indices, self.n_endmembers, self.n_endmembers))
        f_jrs[np.ix_(range(n_indices), indices, indices)] = 1.0

        ival = np.array([indices, exponents]).T
        for k, (i, iexp) in enumerate(ival):
            for l, (j, jexp) in enumerate(ival[k:]):
                if i == j:
                    f_jrs[k, i, i] = iexp * (iexp - 1)
                else:
                    f_jrs[k, i, j] = iexp
                    f_jrs[k, j, i] = iexp
                    f_jrs[l, j, i] = jexp
                    f_jrs[l, i, j] = jexp

        f_rs = np.prod(f_jrs, axis=0)
        return Interaction(indices, exponents, f_r, m_jr, f_rs, m_jrs)

    def _make_interaction_arrays(self, Ws, n):
        """
        A hidden convenience function that splits each
        excess term into _summary_

        :param Ws: List of interactions in form input by the user.
        :type Ws: list of lists
        :param n: Number of interaction values (e.g. 3 for ESV interactions)
        :type n: integer
        :return: The processed interaction values, indices and exponents.
        :rtype: tuple of 2D numpy array and list of Interaction objects
        """
        values = np.empty((len(Ws), n))
        interactions = []
        for i, W in enumerate(Ws):
            values[i] = W[:n]
            interactions.append(self._make_prefactors_and_exponents(W[n:]))
        return values, interactions

    @material_property
    def _c_xs(self):
        """
        Calculates the product of molar fractions raised to the exponents in
        each polynomial interaction.
        :return: Interaction postfactors
            (not including the values of the interactions themselves)
        :rtype: 1D numpy array
        """
        c = np.empty(self.n_W_ESV + self.n_W_mbr)
        for i in range(self.n_W_ESV):
            c[i] = np.prod(
                np.power(self.trans_fractions[self.c_ESV[i].inds], self.c_ESV[i].expts),
                axis=0,
            )
        for i in range(self.n_W_mbr):
            j = i + self.n_W_ESV
            c[j] = np.prod(
                np.power(self.trans_fractions[self.c_mbr[i].inds], self.c_mbr[i].expts),
                axis=0,
            )

        return c

    @material_property
    def _dc_xsdq(self):
        """
        Calculates the first derivative of the product of
        molar fractions raised to the exponents in each polynomial interaction
        with respect to the transformed endmember proportions.
        :return: Interaction postfactors
            (not including the values of the interactions themselves)
        :rtype: 2D numpy array
        """
        c = np.empty((self.n_W_ESV + self.n_W_mbr, self.n_transformed_endmembers))
        for i in range(self.n_W_ESV):
            c[i] = self.c_ESV[i].f_r * np.prod(
                np.power(
                    self.trans_fractions[self.c_ESV[i].inds], self.c_ESV[i].m_jr.T
                ),
                axis=1,
            )
        for i in range(self.n_W_mbr):
            j = i + self.n_W_ESV
            c[j] = self.c_mbr[i].f_r * np.prod(
                np.power(
                    self.trans_fractions[self.c_mbr[i].inds], self.c_mbr[i].m_jr.T
                ),
                axis=1,
            )
        return c

    @material_property
    def _d2c_xsdqdq(self):
        """
        Calculates the second derivative of the product of
        molar fractions raised to the exponents in each polynomial interaction
        with respect to the transformed endmember proportions.
        :return: Interaction postfactors
            (not including the values of the interactions themselves)
        :rtype: 3D numpy array
        """
        c = np.empty(
            (
                self.n_W_ESV + self.n_W_mbr,
                self.n_transformed_endmembers,
                self.n_transformed_endmembers,
            )
        )
        for i in range(self.n_W_ESV):
            c[i] = self.c_ESV[i].f_rs * np.prod(
                np.power(
                    self.trans_fractions[self.c_ESV[i].inds], self.c_ESV[i].m_jrs.T
                ),
                axis=2,
            )
        for i in range(self.n_W_mbr):
            j = i + self.n_W_ESV
            c[j] = self.c_mbr[i].f_rs * np.prod(
                np.power(
                    self.trans_fractions[self.c_mbr[i].inds], self.c_mbr[i].m_jrs.T
                ),
                axis=2,
            )
        return c

    @material_property
    def _dqdn(self):
        """
        The first derivative of the transformed endmember proportions
        with respect to the original endmember amounts.

        :return: dqdn
        :rtype: 2D numpy array
        """
        return np.einsum(
            "ik, km->im",
            self.dqdp,
            dpdn(self.molar_fractions, 1.0, self.ones, self.eye),
        )

    @material_property
    def _dc_xsdn(self):
        """
        Calculates the first derivative of the product of
        molar fractions raised to the exponents in each polynomial interaction
        with respect to the endmember amounts.
        :return: Interaction postfactors
            (not including the values of the interactions themselves)
        :rtype: 2D numpy array
        """
        return np.einsum("ir, rm->im", self._dc_xsdq, self._dqdn) + np.einsum(
            "i, m->im", self._c_xs, self.ones
        )

    @material_property
    def _d2c_xsdndn(self):
        """
        Calculates the second derivative of the product of
        molar fractions raised to the exponents in each polynomial interaction
        with respect to the endmember amounts.
        :return: Interaction postfactors
            (not including the values of the interactions themselves)
        :rtype: 3D numpy array
        """
        n = 1.0
        a1 = np.einsum("irs, sn, rm", self._d2c_xsdqdq, self._dqdn, self._dqdn)
        a2a = 1.0 / n * (np.einsum("ir, rm, n", self._dc_xsdq, self._dqdn, self.ones))
        a2 = a2a + np.einsum("imn->inm", a2a)
        a3 = np.einsum(
            "ir, rk, knm->inm",
            self._dc_xsdq,
            self.dqdp,
            d2pdndn(self.molar_fractions, n * n, self.ones, self.eyeones),
        )
        return a1 + a2 + a3

    @material_property
    def _ESV_gradient_list(self):
        """
        Calculates the first derivative of the
        internal energy, entropy and volume
        with respect to the endmember amounts.

        :return: dEdn, dSdn and dVdn.
        :rtype: 2D numpy array
        """
        if self.W_ESV is None:
            return np.zeros((3, self.n_endmembers))
        else:
            return np.einsum("ij, ik->jk", self.W_ESV, self._dc_xsdn[: self.n_W_ESV])

    @material_property
    def _ESV_hessian_list(self):
        """
        Calculates the second derivative of the
        internal energy, entropy and volume
        with respect to the endmember amounts.

        :return: d2Edndn, d2Sdndn and d2Vdndn.
        :rtype: 3D numpy array
        """
        if self.W_ESV is None:
            return np.zeros((3, self.n_endmembers, self.n_endmembers))
        else:
            return np.einsum(
                "ij, ikl->jkl", self.W_ESV, self._d2c_xsdndn[: self.n_W_ESV]
            )

    @material_property
    def mbr_scalar_list(self):
        """
        :return: The value with which to multiply
        each interaction endmember property to get the
        total excess property.
        :rtype: 1D numpy array
        """
        if self.W_mbr is None:
            return np.zeros((0))
        else:
            return np.einsum("ij, i->j", self.W_mbr, self._c_xs[self.n_W_ESV :])

    @material_property
    def _mbr_gradient_list(self):
        """
        :return: The values with which to multiply
        each interaction endmember property to get the
        first derivatives of the total excess property
        with respect to the endmember amounts.
        :rtype: 2D numpy array
        """
        if self.W_mbr is None:
            return np.zeros((0, self.n_endmembers))
        else:
            return np.einsum("ij, ik->jk", self.W_mbr, self._dc_xsdn[self.n_W_ESV :])

    @material_property
    def _mbr_hessian_list(self):
        """
        :return: The values with which to multiply
        each interaction endmember property to get the
        second derivatives of the total excess property
        with respect to the endmember amounts.
        :rtype: 3D numpy array
        """
        if self.W_mbr is None:
            return np.zeros((0, self.n_endmembers, self.n_endmembers))
        else:
            return np.einsum(
                "ij, ikl->jkl", self.W_mbr, self._d2c_xsdndn[self.n_W_ESV :]
            )

    def _non_ideal_excess_partial_gibbs(self, pressure, temperature, molar_fractions):
        dEdn, dSdn, dVdn = self._ESV_gradient_list
        mbr_gradients = self._mbr_gradient_list
        mbr_gibbs = np.array([mbr.gibbs for mbr in self.interaction_endmembers])
        gibbs = dEdn - temperature * dSdn + pressure * dVdn
        gibbs += np.einsum("i, ij->j", mbr_gibbs, mbr_gradients)
        return gibbs

    def excess_partial_gibbs_free_energies(
        self, pressure, temperature, molar_fractions
    ):
        ideal_gibbs = IdealSolution._ideal_excess_partial_gibbs(
            self, temperature, molar_fractions
        )
        non_ideal_gibbs = self._non_ideal_excess_partial_gibbs(
            pressure, temperature, molar_fractions
        )
        return ideal_gibbs + non_ideal_gibbs

    def excess_partial_entropies(self, pressure, temperature, molar_fractions):
        ideal_entropies = IdealSolution._ideal_excess_partial_entropies(
            self, temperature, molar_fractions
        )

        dSdn = self._ESV_gradient_list[1]
        mbr_gradients = self._mbr_gradient_list
        mbr_entropies = np.array([mbr.S for mbr in self.interaction_endmembers])

        non_ideal_entropies = dSdn + np.einsum("i, ij->j", mbr_entropies, mbr_gradients)
        return ideal_entropies + non_ideal_entropies

    def excess_partial_volumes(self, pressure, temperature, molar_fractions):
        dVdn = self._ESV_gradient_list[2]
        mbr_gradients = self._mbr_gradient_list
        mbr_volumes = np.array([mbr.V for mbr in self.interaction_endmembers])

        return dVdn + np.einsum("i, ij->j", mbr_volumes, mbr_gradients)

    def gibbs_hessian(self, pressure, temperature, molar_fractions):
        ideal_entropy_hessian = IdealSolution._ideal_entropy_hessian(
            self, temperature, molar_fractions
        )

        d2Edndn, d2Sdndn, d2Vdndn = self._ESV_hessian_list
        mbr_hessian = self._mbr_hessian_list
        mbr_gibbs = np.array([mbr.gibbs for mbr in self.interaction_endmembers])

        d2Gdndn = d2Edndn - temperature * d2Sdndn + pressure * d2Vdndn
        d2Gdndn += np.einsum("i, ijk->jk", mbr_gibbs, mbr_hessian)

        return d2Gdndn - temperature * ideal_entropy_hessian

    def entropy_hessian(self, pressure, temperature, molar_fractions):
        ideal_entropy_hessian = IdealSolution._ideal_entropy_hessian(
            self, temperature, molar_fractions
        )

        d2Sdndn = self._ESV_hessian_list[1]
        mbr_hessian = self._mbr_hessian_list
        mbr_entropies = np.array([mbr.S for mbr in self.interaction_endmembers])

        d2Sdndn += np.einsum("i, ijk->jk", mbr_entropies, mbr_hessian)

        return d2Sdndn + ideal_entropy_hessian

    def volume_hessian(self, pressure, temperature, molar_fractions):
        d2Vdndn = self._ESV_hessian_list[2]
        mbr_hessian = self._mbr_hessian_list
        mbr_volumes = np.array([mbr.V for mbr in self.interaction_endmembers])

        d2Vdndn += np.einsum("i, ijk->jk", mbr_volumes, mbr_hessian)

        return d2Vdndn

    def Cp_excess(self):
        mbr_scalar = self.mbr_scalar_list
        mbr_Cp = np.array(
            [mbr.molar_heat_capacity_p for mbr in self.interaction_endmembers]
        )
        return np.einsum("i, i", mbr_scalar, mbr_Cp)

    def alphaV_excess(self):
        mbr_scalar = self.mbr_scalar_list
        mbr_d2gibbsdpdt = np.array(
            [mbr.alpha * mbr.V for mbr in self.interaction_endmembers]
        )
        return np.einsum("i, i", mbr_scalar, mbr_d2gibbsdpdt)

    def VoverKT_excess(self):
        mbr_scalar = self.mbr_scalar_list
        mbr_d2gibbsdpdp = np.array(
            [
                mbr.V / mbr.isothermal_bulk_modulus_reuss
                for mbr in self.interaction_endmembers
            ]
        )
        return np.einsum("i, i", mbr_scalar, mbr_d2gibbsdpdp)

    def activity_coefficients(self, pressure, temperature, molar_fractions):
        if temperature > 1.0e-10:
            return np.exp(
                self._non_ideal_excess_partial_gibbs(
                    pressure, temperature, molar_fractions
                )
                / (constants.gas_constant * temperature)
            )
        else:
            raise Exception("Activity coefficients not defined at 0 K.")

    def activities(self, pressure, temperature, molar_fractions):
        return IdealSolution.activities(
            self, pressure, temperature, molar_fractions
        ) * self.activity_coefficients(pressure, temperature, molar_fractions)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
from Kurnosov et al.  Nature 2017
Kurnosov, A., et al. "Evidence for a Fe3+-rich pyrolitic lower mantle from
(Al, Fe)-bearing bridgmanite elasticity data."
Nature 543.7646 (2017): 543-546. doi:10.1038/nature21390
"""

from __future__ import absolute_import

from ..classes.mineral import Mineral
from ..classes.solution import Solution
from ..classes.solutionmodel import IdealSolution, SymmetricRegularSolution
from ..utils.chemistry import dictionarize_formula, formula_mass

"""
SOLID SOLUTIONS
"""


class bridgmanite(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "bridgmanite/perovskite"
        self.solution_model = IdealSolution(
            endmembers=[
                [mg_si_perovskite(), "[Mg][Si]O3"],
                [fe_si_perovskite(), "[Fe][Si]O3"],
                [al_al_perovskite(), "[Al][Al]O3"],
                [fe_al_perovskite(), "[Fe][Al]O3"],
            ]
        )
        Solution.__init__(self, molar_fractions=molar_fractions)


class ferropericlase(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "magnesiowustite/ferropericlase"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[[periclase(), "[Mg]O"], [wuestite(), "[Fe]O"]],
            energy_interaction=[[13.0e3]],
        )

        Solution.__init__(self, molar_fractions=molar_fractions)


"""
ENDMEMBERS
"""


class mg_si_perovskite(Mineral):
    def __init__(self):
        formula = "MgSiO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "MgSiO3 perovskite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1368000.0,
            "V_0": 2.4445e-05,
            "K_0": 2.51e11,
            "Kprime_0": 4.1,
            "Debye_0": 905.0,
            "grueneisen_0": 1.57,
            "q_0": 1.1,
            "G_0": 1.73e11,
            "Gprime_0": 1.7,
            "eta_s_0": 2.3,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        Mineral.__init__(self)


class fe_si_perovskite(Mineral):
    def __init__(self):
        formula = "FeSiO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "FeSiO3 perovskite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1043000.0,
            "V_0": 2.534e-05,
            "K_0": 2.72e11,
            "Kprime_0": 4.1,
            "Debye_0": 871.0,
            "grueneisen_0": 1.57,
            "q_0": 1.1,
            "G_0": 1.33 + 11,
            "Gprime_0": 1.4,
            "eta_s_0": 2.3,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        Mineral.__init__(self)


class fe_al_perovskite(Mineral):
    def __init__(self):
        formula = "FeAlO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "FeAlO3 perovskite",
            "formula": formula,
            "equation_of_state": "slb3",
            "V_0": 2.69e-05,
            "K_0": 2.20e11,
            "Kprime_0": 1.3,
            "Debye_0": 886.0,
            "grueneisen_0": 1.57,
            "q_0": 1.1,
            "G_0": 96.0e09,
            "Gprime_0": 3.4,
            "eta_s_0": 2.5,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        Mineral.__init__(self)


class al_al_perovskite(Mineral):
    def __init__(self):
        formula = "Al2O3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Al2O3 perovskite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1533878.0,
            "V_0": 2.494e-05,
            "K_0": 2.58e11,
            "Kprime_0": 4.1,
            "Debye_0": 886.0,
            "grueneisen_0": 1.57,
            "q_0": 1.1,
            "G_0": 1.71e11,
            "Gprime_0": 1.5,
            "eta_s_0": 2.5,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        Mineral.__init__(self)


class periclase(Mineral):
    def __init__(self):
        formula = "MgO"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Periclase",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -569000.0,
            "V_0": 1.124e-05,
            "K_0": 1.602e11,
            "Kprime_0": 3.99,
            "Debye_0": 767.0,
            "grueneisen_0": 1.36,
            "q_0": 1.7,
            "G_0": 1.31e11,
            "Gprime_0": 2.1,
            "eta_s_0": 2.8,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        Mineral.__init__(self)


class wuestite(Mineral):
    def __init__(self):
        formula = "FeO"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Wuestite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -242000.0,
            "V_0": 1.226e-05,
            "K_0": 1.49e11,
            "Kprime_0": 3.6,
            "Debye_0": 454.0,
            "grueneisen_0": 1.53,
            "q_0": 1.7,
            "G_0": 60.0e9,
            "Gprime_0": 1.8,
            "eta_s_0": 0.6,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        Mineral.__init__(self)


class ca_perovskite(Mineral):
    def __init__(self):
        formula = "CaSiO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "CaSiO3 perovskite",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1463358.0,
            "V_0": 2.754e-05,
            "K_0": 2.36e11,
            "Kprime_0": 3.9,
            "Debye_0": 802.0,
            "grueneisen_0": 1.89,
            "q_0": 0.9,
            "G_0": 1.57e11,
            "Gprime_0": 2.2,
            "eta_s_0": 1.3,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        Mineral.__init__(self)


class hcp_iron(Mineral):
    def __init__(self):
        formula = "Fe"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "Hexagonal close packed iron",
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": -1463358.0,
            "V_0": 0.673e-05,
            "K_0": 1.64e11,
            "Kprime_0": 4.0,
            "Debye_0": 422.0,
            "grueneisen_0": 1.71,
            "q_0": 1.4,
            "G_0": 0.815e11,
            "Gprime_0": 1.9,
            "eta_s_0": 7,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        Mineral.__init__(self)


"""
Mineral aliases
"""
perovskite = bridgmanite



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2015 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
from __future__ import print_function

import numpy as np
from scipy.stats import t, norm, genextreme
import copy

from ..utils.math import unit_normalize
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from matplotlib.patches import Ellipse


def nonlinear_least_squares_fit(
    model, lm_damping=0.0, param_tolerance=1.0e-7, max_lm_iterations=100, verbose=False
):
    """
    Function to compute the "best-fit" parameters for a model
    by nonlinear least squares fitting.

    The nonlinear least squares algorithm closely follows the logic in
    Section 23.1 of Bayesian Probability Theory
    (von der Linden et al., 2014; Cambridge University Press).

    Parameters
    ----------
    :param model: Model containing data to be fit, and functions to
        aid in fitting.
    :type model: object

    :param lm_damping: Levenberg-Marquardt parameter for least squares minimization.
    :type lm_damping: float

    :param param_tolerance: Levenberg-Marquardt iterations are terminated when
        the maximum fractional change in any of the parameters
        during an iteration drops below this value
    :type param_tolerance: float


    :param max_lm_iterations: Maximum number of Levenberg-Marquardt iterations
    :type max_lm_iterations: int

    :param verbose: Print some information to standard output
    :type verbose: bool

    .. note:: The object passed as model must have the following attributes:
        * data [2D numpy.array] - Elements of x[i][j] contain the
            observed position of data point i.
        * data_covariances [3D numpy.array] Elements of cov[i][j][k] contain
            the covariance matrix of data point i.
        * mle_tolerances [numpy.array] - The iterations to find the maximum likelihood
            estimator for each observed data point will stop when mle_tolerances[i] <
            np.linalg.norm(data_mle[i] - model.function(data_mle[i], flag))
        * delta_params [numpy.array] - parameter perturbations used to compute the jacobian

        Must also have the following methods:
        * set_params(self, param_values) -  Function to set parameters.
        * get_params(self) - Function to get current model parameters.
        * function(self, x) - Returns value of model function evaluated at x.
        * normal(self, x) - Returns value of normal to the model function evaluated at x.

        After this function has been performed, the following attributes are added to model:

        * n_dof [int] - Degrees of freedom of the system.
        * data_mle [2D numpy array] - Maximum likelihood estimates of the observed data points
            on the best-fit curve.
        * jacobian [2D numpy array] - d(weighted_residuals)/d(parameter).
        * weighted_residuals [numpy array] - Weighted residuals.
        * weights [numpy array] - 1/(data variances normal to the best fit curve).
        * WSS [float] - Weighted sum of squares residuals.
        * popt [numpy array] - Optimized parameters.
        * pcov [2D numpy array] - Covariance matrix of optimized parameters.
        * noise_variance [float] - Estimate of the variance of the data normal to the curve.

    This function is available as ``burnman.nonlinear_least_squares_fit``.
    """

    def _mle_estimate(x, x_m, cov, flag):
        n = model.normal(x_m, flag)
        var_n = abs_line_project(cov, n)
        d = (x_m - x).dot(n)
        x_mle = x + d * ((n.dot(cov)).T) / var_n
        return x_mle, d, var_n

    def _find_mle():
        x_mle_arr = np.empty_like(model.data)
        residual_arr = np.empty(n_data)
        var_arr = np.empty(n_data)
        for i, (x, cov, flag) in enumerate(
            zip(*[model.data, model.data_covariances, model.flags])
        ):
            x_mle_arr[i] = model.function(x, flag)
            x_mle_est, residual_arr[i], var_arr[i] = _mle_estimate(
                x, x_mle_arr[i], cov, flag
            )
            delta_x = x_mle_arr[i] - x

            while np.linalg.norm(delta_x) > model.mle_tolerances[i]:
                x_mle_est, residual_arr[i], var_arr[i] = _mle_estimate(
                    x, x_mle_arr[i], cov, flag
                )
                x_mle_arr[i] = model.function(x_mle_est, flag)
                delta_x = x_mle_arr[i] - x_mle_est

        return x_mle_arr, residual_arr / np.sqrt(var_arr), 1.0 / var_arr

    def calculate_jacobian():
        model.jacobian = np.empty((n_data, n_params))
        diag_delta = np.diag(model.delta_params)
        param_values = model.get_params()
        for prm_i, value in enumerate(param_values):
            model.set_params(param_values - diag_delta[prm_i])
            x_mle_arr, residual_arr_0, weights_0 = _find_mle()

            model.set_params(param_values + diag_delta[prm_i])
            x_mle_arr, residual_arr_1, weights_1 = _find_mle()

            model.jacobian[:, prm_i] = (residual_arr_1 - residual_arr_0) / (
                2.0 * diag_delta[prm_i][prm_i]
            )
        model.set_params(param_values)  # reset params

    def _update_beta(lmbda):
        # Performs a Levenberg-Marquardt iteration
        # Note that if lambda = 0, this is a simple Gauss-Newton iteration
        calculate_jacobian()
        model.data_mle, model.weighted_residuals, model.weights = _find_mle()

        J = model.jacobian  # this the weighted Jacobian
        JTJ = J.T.dot(J)
        delta_beta = (
            np.linalg.inv(JTJ + lmbda * np.diag(JTJ))
            .dot(J.T)
            .dot(model.weighted_residuals)
        )
        old_params = np.copy(model.get_params())
        new_params = old_params - delta_beta
        # f_delta_beta = delta_beta/new_params

        model.set_params(new_params)

        # set_params may modify the step to satisfy bounds on the problem
        # We therefore need to get the params before
        # calculating the fractional change.
        new_params = model.get_params()

        # In case the new_params object returns a very small value,
        # modify to avoid a pointless comparison:
        mod_params = np.where(
            np.abs(new_params) < param_tolerance, param_tolerance, new_params
        )
        f_delta_beta = (old_params - new_params) / mod_params
        return f_delta_beta

    n_data = len(model.data)
    params = model.get_params()
    n_params = len(params)
    model.dof = n_data - n_params

    if not hasattr(model, "flags"):
        model.flags = [None] * n_data

    for n_it in range(max_lm_iterations):
        # update the parameters with a LM iteration
        f_delta_beta = _update_beta(lm_damping)
        max_f = np.max(np.abs(f_delta_beta))
        if verbose is True:
            print(
                "Iteration {0:d}: {1}. Max change in param: {2}".format(
                    n_it, model.get_params(), max_f
                )
            )
        if max_f < param_tolerance:
            break

    J = model.jacobian
    r = model.weighted_residuals
    model.WSS = r.dot(r.T)

    model.popt = model.get_params()
    model.pcov = np.linalg.inv(J.T.dot(J)) * r.dot(r.T) / model.dof

    # Estimate the noise variance normal to the curve
    model.goodness_of_fit = model.WSS / model.dof
    model.noise_variance = r.dot(np.diag(1.0 / model.weights)).dot(r.T) / model.dof

    if verbose is True:
        if n_it == max_lm_iterations - 1:
            print(
                f"Max iterations ({max_lm_iterations:d}) reached "
                f"(param tolerance = {param_tolerance:1e})"
            )
        else:
            print("Converged in {0:d} iterations".format(n_it))
        print("\nOptimised parameter values:")
        print(model.popt)
        print("\nParameter covariance matrix:")
        print(model.pcov)
        print("")


def confidence_prediction_bands(model, x_array, confidence_interval, f, flag=None):
    """
    This function calculates the confidence and prediction bands of
    the function f(x) from a best-fit model with uncertainties in its
    parameters as calculated (for example) by
    the function nonlinear_least_squares_fit().

    The values are calculated via the delta method, which estimates
    the variance of f evaluated at x as var(f(x)) = df(x)/dB var(B) df(x)/dB
    where df(x)/dB is the vector of partial derivatives of f(x)
    with respect to B.

    :param model: As modified (for example) by the function
        :func:`burnman.nonlinear_least_squares_fit`.
        Should contain the following functions: get_params, set_params, function, normal
        And attributes: delta_params, pcov, dof, noise_variance
    :type model: object

    :param x_array: Coordinates at which to evaluate the bounds.
    :type x_array: 2D numpy.array

    :param confidence_interval: Probability level of finding the true model
        (confidence bound) or any new data point (probability bound).
        For example, the 95% confidence bounds should be calculated using a
        confidence interval of 0.95.
    :type confidence_interval: float

    :param f: The function defining the variable y=f(x) for which the
        confidence and prediction bounds are desired.
    :type f: function

    :param flag: This (optional) flag is passed to model.function to control how the
        modified position of x is calculated. This value is then used by f(x)
    :type flag: type informed by model object

    :returns: An element of bounds[i][j] gives the lower and upper confidence
        (i=0, i=1) and prediction (i=2, i=3) bounds for the jth data point.
    :rtype: 2D numpy.array
    """

    # Check array dimensions
    n_dimensions = len(model.data[0])
    if len(x_array[0]) != n_dimensions:
        raise Exception(
            "Dimensions of each point must be the same as the "
            "total number of dimensions"
        )

    param_values = model.get_params()
    x_m_0s = np.empty_like(x_array)
    f_m_0s = np.empty_like(x_array[:, 0])
    for i, x in enumerate(x_array):
        x_m_0s[i] = model.function(x, flag)
        f_m_0s[i] = f(x)

    diag_delta = np.diag(model.delta_params)
    dxdbeta = np.empty([len(param_values), len(x_array)])

    for i, value in enumerate(param_values):
        model.set_params(param_values + diag_delta[i])

        for j, x_m_0 in enumerate(x_m_0s):
            x_m_1 = model.function(x_m_0, flag)
            dxdbeta[i][j] = (f(x_m_1) - f_m_0s[j]) / diag_delta[i][i]

    model.set_params(param_values)  # reset params

    variance = np.empty(len(x_array))
    for i, Gprime in enumerate(dxdbeta.T):
        variance[i] = Gprime.T.dot(model.pcov).dot(Gprime)

    critical_value = t.isf(0.5 * (confidence_interval + 1.0), model.dof)

    confidence_half_widths = critical_value * np.sqrt(variance)
    prediction_half_widths = critical_value * np.sqrt(variance + model.noise_variance)

    confidence_bound_0 = f_m_0s - confidence_half_widths
    confidence_bound_1 = f_m_0s + confidence_half_widths
    prediction_bound_0 = f_m_0s - prediction_half_widths
    prediction_bound_1 = f_m_0s + prediction_half_widths

    return np.array(
        [confidence_bound_0, confidence_bound_1, prediction_bound_0, prediction_bound_1]
    )


def abs_line_project(M, n):
    n = unit_normalize(n)
    return n.dot(M).dot(n.T)


def plot_cov_ellipse(cov, pos, nstd=2, ax=None, **kwargs):
    """
    Plots an `nstd` sigma error ellipse based on the specified covariance
    matrix (`cov`). Additional keyword arguments are passed on to the
    ellipse patch artist.

    :param cov: The 2x2 covariance matrix to base the ellipse on.
    :type cov: numpy.array

    :param pos: The location of the center of the ellipse. Expects a 2-element
        sequence of [x0, y0].
    :type pos: list or numpy.array

    :param nstd: The radius of the ellipse in numbers of standard deviations.
        Defaults to 2 standard deviations.
    :type nstd: float

    :param ax: The axis that the ellipse will be plotted on. Defaults to the
        current axis.
    :type ax: matplotlib.pyplot.axes

    :param kwargs: Additional keyword arguments are passed on to the ellipse patch.

    :returns: The covariance ellipse (already applied to the desired axes object).
    :rtype: matplotlib.patches.Ellipse
    """

    def eigsorted(cov):
        vals, vecs = np.linalg.eigh(cov)
        order = vals.argsort()[::-1]
        return vals[order], vecs[:, order]

    if ax is None:
        ax = plt.gca()

    vals, vecs = eigsorted(cov)
    theta = np.degrees(np.arctan2(*vecs[:, 0][::-1]))

    # Width and height are "full" widths, not radius
    width, height = 2 * nstd * np.sqrt(vals)
    ellip = Ellipse(xy=pos, width=width, height=height, angle=theta, **kwargs)

    ax.add_artist(ellip)
    return ellip


def corner_plot(popt, pcov, param_names=[], n_std=1.0):
    """
    Creates a corner plot of covariances

    :param popt: Optimized parameters.
    :type popt: numpy.array

    :param pcov: Covariance matrix of the parameters.
    :type pcov: 2D numpy.array

    :param param_names: Parameter names.
    :type param_names: list

    :param n_std: Number of standard deviations for ellipse.
    :type n_std: float

    :returns: ``matplotlib.pyplot.figure`` and list of ``matplotlib.pyplot.Axes``
        objects.
    :rtype: tuple
    """

    if len(pcov[0]) != len(pcov[:, 0]):
        raise Exception("Covariance matrices must be square")

    n_params = len(pcov[0])
    if n_params < 2:
        raise Exception(
            "Covariance matrix must be at least 2x2 for " "a corner plot to be plotted"
        )

    # ellipse plotting is prone to rounding errors, so we scale the plots here
    scaling = 1.0 / np.power(10.0, np.around(np.log10(np.abs(popt)) - 0.5))
    scaling = np.outer(scaling, scaling)

    fig, ax = plt.subplots(n_params - 1, n_params - 1)
    fig.set_size_inches(3.0 * (n_params - 1), 3.0 * (n_params - 1))

    for j in range(n_params - 1):
        for i in range(j):
            fig.delaxes(ax[i][j])

        for i in range(j, n_params - 1):
            ax[i][j].get_xaxis().get_major_formatter().set_useOffset(False)
            ax[i][j].get_yaxis().get_major_formatter().set_useOffset(False)
            ax[i][j].set_box_aspect(1)

            if j > 0:
                ax[i][j].get_yaxis().set_visible(False)
            if i < n_params - 2:
                ax[i][j].get_xaxis().set_visible(False)

            indices = np.array([j, i + 1])
            projected_cov = (pcov * scaling)[indices[:, None], indices]

            scaled_pos = np.array(
                [
                    popt[j] * np.sqrt(scaling[j][j]),
                    popt[i + 1] * np.sqrt(scaling[i + 1][i + 1]),
                ]
            )

            plot_cov_ellipse(
                cov=projected_cov, pos=scaled_pos, nstd=n_std, ax=ax[i][j], color="grey"
            )
            maxx = 1.5 * n_std * np.sqrt(projected_cov[0][0])
            maxy = 1.5 * n_std * np.sqrt(projected_cov[1][1])
            ax[i][j].set_xlim(scaled_pos[0] - maxx, scaled_pos[0] + maxx)
            ax[i][j].set_ylim(scaled_pos[1] - maxy, scaled_pos[1] + maxy)

    if param_names != []:
        for i in range(n_params - 1):
            ax[n_params - 2][i].set_xlabel(
                "{0:s} (x $10^{{{1:d}}}$)".format(
                    param_names[i], -int(np.log10(np.sqrt(scaling[i][i])))
                )
            )

        for j in range(1, n_params):
            ax[j - 1][0].set_ylabel(
                "{0:s} (x $10^{{{1:d}}}$)".format(
                    param_names[j], -int(np.log10(np.sqrt(scaling[j][j])))
                )
            )

    fig.set_tight_layout(True)

    return fig, ax


def weighted_residual_plot(
    ax,
    model,
    flag=None,
    sd_limit=3,
    cmap=plt.cm.RdYlBu,
    plot_axes=[0, 1],
    scale_axes=[1.0, 1.0],
):
    """
    Creates a plot of the weighted residuals
    The user can choose the projection axes, and scaling to apply to those axes
    The chosen color palette (cmap) is discretised by standard deviation up
    to a cut off value of sd_limit.

    :param ax: Plot.
    :param type: ``matplotlib.pyplot.Axes``

    :param model: A model as used by
        :func:`burnman.nonlinear_least_squares_fit`.
        Must contain the attributes model.data,
        model.weighted_residuals and
        model.flags (if flag is not None).
    :type model: object

    :param flag: String to determine which data to plot.
        Finds matches with model.flags.
    :type flag: str

    :param sd_limit: Data with weighted residuals exceeding this
        limit are plotted in black.
    :type sd_limit: float

    :param cmap: Color palette.
    :type cmap: matplotlib color palette

    :param plot_axes: Data axes to use as plot axes.
    :type plot_axes: list of int

    :param scale_axes: Plot axes are scaled by multiplication
        of the data by these values.
    :type scale_axes: list of float

    :returns: Coloured scatter plot of the weighted residuals in data space.
    :rtype: matplotlib Axes object
    """
    if flag is None:
        mask = range(len(model.data[:, 0]))
    else:
        mask = [i for i, flg in enumerate(model.flags) if flg == flag]

    cmap_cp = copy.copy(cmap)
    cmap_cp.set_under("k")
    cmap_cp.set_over("k")
    bounds = np.linspace(-sd_limit, sd_limit, sd_limit * 2 + 1)
    norm = colors.BoundaryNorm(bounds, cmap_cp.N)

    im = ax.scatter(
        model.data[:, plot_axes[0]][mask] * scale_axes[0],
        model.data[:, plot_axes[1]][mask] * scale_axes[1],
        c=model.weighted_residuals[mask],
        cmap=cmap_cp,
        norm=norm,
        s=50,
    )
    plt.colorbar(im, ax=ax, label="Misfit (standard deviations)")


def extreme_values(weighted_residuals, confidence_interval):
    """
    This function uses extreme value theory to calculate the number of
    standard deviations away from the mean at which we should expect to bracket
    *all* of our n data points at a certain confidence level.

    It then uses that value to identify which (if any) of the data points
    lie outside that region, and calculates the corresponding probabilities
    of finding a data point at least that many standard deviations away.


    :param weighted_residuals: Array of residuals weighted by the square root
        of their variances wr_i = r_i/sqrt(var_i).
    :type weighted_residuals: array of float

    :param confidence_interval: Probability at which all the weighted residuals lie
        within the confidence bounds.
    :type confidence_interval: float

    :returns: Number of standard deviations at which we should expect to encompass
        all data at the user-defined confidence interval, the indices of weighted
        residuals exceeding the confidence_interval defined by the user, and
        the probabilities that the extreme data point of the distribution lies
        further from the mean than the observed position wr_i for each i in
        the "indices" output array.
    :rtype: tuple of (float, numpy.array, numpy.array)
    """

    n = len(weighted_residuals)
    mean = norm.isf(1.0 / n)
    # good approximation for > 10 data points
    scale = 0.8 / np.power(np.log(n), 1.0 / 2.0)
    # good approximation for > 10 data points
    c = 0.33 / np.power(np.log(n), 3.0 / 4.0)

    # We now need a 1-tailed probability from the given confidence_interval
    # p_total = 1. - confidence_interval = p_upper + p_lower - p_upper*p_lower
    # p_total = 1. - confidence_interval = 2p - p^2, therefore:
    p = 1.0 - np.sqrt(confidence_interval)
    confidence_bound = genextreme.isf(p, c, loc=mean, scale=scale)

    indices = [
        i for i, r in enumerate(weighted_residuals) if np.abs(r) > confidence_bound
    ]
    # Convert back to 2-tailed probabilities
    probabilities = 1.0 - np.power(
        genextreme.sf(np.abs(weighted_residuals[indices]), c, loc=mean, scale=scale)
        - 1.0,
        2.0,
    )

    return confidence_bound, indices, probabilities


def plot_residuals(ax, weighted_residuals, n_bins=None, flags=[]):
    if flags is []:
        flags = [""] * len(weighted_residuals)
        list_flags = [""]
    else:
        list_flags = list(set(flags))

    if n_bins is None:
        try:  # Only works for recent versions of numpy
            bin_heights, bin_bounds = np.histogram(
                weighted_residuals, bins="auto", density=True
            )
            n_bins = len(bin_heights)
        except:
            n_bins = 11.0

    mask = [i for i, f in enumerate(flags)]
    for flag in list_flags:
        binwidth = np.ptp(weighted_residuals) / n_bins
        dmin = min(weighted_residuals) - binwidth
        dmax = max(weighted_residuals) + binwidth
        bins = np.linspace(dmin, dmax, n_bins)
        bin_heights, bin_bounds = np.histogram(
            weighted_residuals[mask], bins=bins, density=True
        )

        normalisation = float(len(weighted_residuals[mask])) / float(
            len(weighted_residuals)
        )
        bin_centers = (bin_bounds[:-1] + bin_bounds[1:]) / 2.0
        bin_heights = bin_heights * normalisation
        bin_widths = bin_bounds[1] - bin_bounds[0]
        plt.bar(bin_centers, bin_heights, width=bin_widths, label=flag, alpha=0.2)
        mask = [i for i, f in enumerate(flags) if f != flag and i in mask]

        x = np.linspace(bin_bounds[0], bin_bounds[-1], 1001)
        ax.plot(x, norm.pdf(x) * normalisation)

    ax.set_title("Residual plot versus expected normal distribution")
    ax.set_xlabel("Number of standard deviations from the mean")
    ax.set_ylabel("Probability")
    ax.legend(loc="upper right")
